try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import ezdxf
except BaseException:
    ezdxf = None


class DXFTest(g.unittest.TestCase):
    def test_dxf(self):
        # get a path we can write
        temp_name = g.tempfile.NamedTemporaryFile(suffix=".dxf", delete=False).name
        loaded = g.get_2D()
        # split drawings into single body parts
        splits = []
        for d in loaded:
            s = d.split()
            # check area of split result vs source
            assert g.np.isclose(sum(i.area for i in s), d.area)
            splits.append(s)

            # export the drawing to the file
            d.export(file_obj=temp_name)

            # try using ezdxf as a simple validator
            # it raises exceptions aggressively
            if ezdxf is not None:
                with open(temp_name) as f:
                    ezdxf.read(f)

            # export to a string
            text = d.export(file_type="dxf")

            # DXF files are always pairs of lines
            lines = str.splitlines(str(text))
            assert (len(lines) % 2) == 0
            assert all(len(L.strip()) > 0 for L in lines)

            # reload the file by name and by stream
            rc = [
                g.trimesh.load(temp_name),
                g.trimesh.load(g.io_wrap(text), file_type="dxf"),
            ]

            # compare reloaded with original
            for r in rc:
                assert g.np.isclose(r.area, d.area)
                assert g.np.isclose(r.length, d.length, rtol=1e-4)
                assert len(r.entities) == len(d.entities)

        single = g.np.hstack(splits)
        for p in single:
            p.vertices /= p.scale

            # make sure exporting by name works
            # use tempfile to avoid dumping file in
            # our working directory
            p.export(temp_name)
            r = g.trimesh.load(temp_name)

            ratio = abs(p.length - r.length) / p.length
            if ratio > 0.01:
                g.log.error(
                    "perimeter ratio on export %s wrong! %f %f %f",
                    p.source.file_name,
                    p.length,
                    r.length,
                    ratio,
                )

                raise ValueError(
                    f"perimeter ratio too large ({ratio}) on {p.source.file_name}"
                )

    def test_spline(self):
        d = g.get_mesh("2D/cycloidal.dxf")

        assert len(d.entities) == 1
        assert type(d.entities[0]).__name__ == "BSpline"

        # export to dxf and wrap as a file object
        e = g.trimesh.util.wrap_as_stream(d.export(file_type="dxf"))
        # reconstitute drawing
        r = g.trimesh.load(e, file_type="dxf")

        # make sure reconstituted drawing is the same as the source
        assert len(r.entities) == 1
        assert type(r.entities[0]).__name__ == "BSpline"
        assert g.np.isclose(r.area, d.area)

        assert len(d.entities[0].points) == len(r.entities[0].points)
        assert len(d.entities[0].knots) == len(r.entities[0].knots)

    def test_versions(self):
        """
        DXF files have a bajillion versions, so test against
        the same files saved in multiple versions by 2D CAD
        packages.

        Version test files are named things like:
        ae.r14a.dxf: all entity types, R14 ASCII DXF
        uc.2007b.dxf: unit square, R2007 binary DXF
        """
        # directory where multiple versions of DXF are
        dir_versions = g.os.path.join(g.dir_2D, "versions")

        # load the different versions
        paths = {}
        for f in g.os.listdir(dir_versions):
            # full path including directory
            ff = g.os.path.join(dir_versions, f)
            try:
                paths[f] = g.trimesh.load(ff)
            except NotImplementedError as E:
                # something like 'r14a' for ascii
                # and 'r14b' for binary
                version = f.split(".")[-2]
                # we should only get ValueErrors on binary DXF
                assert version[-1] == "b"
                g.log.debug(E, f)

        # group drawings which have the same geometry
        # but exported in different revisions of the DXF format
        groups = g.collections.defaultdict(list)
        for k in paths.keys():
            # the first string before a period is the drawing name
            groups[k.split(".")[0]].append(k)

        # loop through each group of the same drawing
        for group in groups.values():
            # get the total length of every entity
            L = [paths[i].length for i in group]
            L = g.np.array(L, dtype=g.np.float64)

            # make sure all versions have consistent length
            assert g.np.allclose(L, L.mean(), rtol=0.01)

            # count the number of entities in the path
            # this should be the same for every version
            E = g.np.array([len(paths[i].entities) for i in group], dtype=g.np.int64)
            assert g.np.ptp(E) == 0

    def test_bulge(self):
        """
        Test bulged polylines which are polylines with
        implicit arcs.
        """
        # get a drawing with bulged polylines
        p = g.get_mesh("2D/LM2.dxf")
        # count the number of unclosed arc entities
        # this drawing only has polylines with bulge
        spans = [
            e.center(p.vertices)["span"]
            for e in p.entities
            if type(e).__name__ == "Arc" and not e.closed
        ]
        # should have only one outer loop
        assert len(p.root) == 1
        # should have 6 partial arcs from bulge
        assert len(spans) == 6
        # all arcs should be 180 degree slot end caps
        assert g.np.allclose(spans, g.np.pi)

    def test_text(self):
        # load file with a single text entity
        original = g.get_mesh("2D/text.dxf")

        # export then reload
        roundtrip = g.trimesh.load(
            file_obj=g.io_wrap(original.export(file_type="dxf")), file_type="dxf"
        )

        for d in [original, roundtrip]:
            # should contain a single Text entity
            assert len(d.entities) == 1

            # shouldn't crash anything
            assert len(d.polygons_closed) == 0
            assert len(d.polygons_full) == 0
            assert len(d.discrete) == 0
            assert len(d.paths) == 0

            # make sure it preserved case and special chars
            assert d.entities[0].text == "HEY WHAT's poppin"

            # height should 1.0
            assert g.np.isclose(d.entities[0].height, 1.0)

            # get the 2D rotation of the text
            angle = d.entities[0].angle(d.vertices)

            # angle should be 30 degrees
            assert g.np.isclose(angle, g.np.radians(30.0))

    def test_unicode(self):
        """
        Check our handling of unicode. Current approach is to
        just force everything into ASCII rather than handling
        the encoding flags in DXF headers.
        """
        # get a base 2D model
        m = g.get_mesh("2D/wrench.dxf")
        # make one of the entity layers a unicode string
        # store it as B64 so python2 doesn't get mad
        layer = g.base64.b64decode("VFJBw4dBRE9IT1JJWk9OVEFMX1RSQcOHQURPNA==").decode(
            "utf-8"
        )
        m.entities[0].layer = layer
        # export to a string
        export = m.export(file_type="dxf")
        # if any unicode survived the export this will fail
        export.encode("ascii")

    def test_insert_block(self):
        a = g.get_mesh("2D/insert.dxf")
        b = g.get_mesh("2D/insert_r14.dxf")

        assert len(a.polygons_full) == 2
        assert len(b.polygons_full) == 2

        assert g.np.isclose(a.area, 54075.0, atol=1)
        assert g.np.isclose(b.area, 54075.0, atol=1)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



from functools import reduce

import numpy as np

from trimesh.iteration import chain, reduce_cascade


def test_reduce_cascade():
    # the multiply will explode quickly past the integer maximum
    def both(operation, items):
        """
        Run our cascaded reduce and regular reduce.
        """

        b = reduce_cascade(operation, items)

        if len(items) > 0:
            assert b == reduce(operation, items)

        return b

    for i in range(20):
        data = np.arange(i)
        c = both(items=data, operation=lambda a, b: a + b)

        if i == 0:
            assert c is None
        else:
            assert c == np.arange(i).sum()

        # try a multiply
        data = np.arange(i)
        c = both(items=data, operation=lambda a, b: a * b)

        if i == 0:
            assert c is None
        else:
            assert c == np.prod(data)

        # try a multiply
        data = np.arange(i)[1:]
        c = both(items=data, operation=lambda a, b: a * b)
        if i <= 1:
            assert c is None
        else:
            assert c == np.prod(data)

    data = ["a", "b", "c", "d", "e", "f", "g"]
    print("# reduce_pairwise\n-----------")
    r = both(operation=lambda a, b: a + b, items=data)

    assert r == "abcdefg"


def test_chain():
    # should work on iterables the same as `itertools.chain`
    assert np.allclose(chain([1, 3], [4]), [1, 3, 4])
    # should work with non-iterable single values
    assert np.allclose(chain([1, 3], 4), [1, 3, 4])
    # should filter out `None` arguments
    assert np.allclose(chain([1, 3], None, 4, None), [1, 3, 4])


if __name__ == "__main__":
    test_reduce_cascade()



import numpy as np


def test_resample_original():
    # check to see if `include_original` works

    from shapely.geometry import Polygon

    from trimesh.path.traversal import resample_path

    ori = np.array([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]], dtype=np.float64)

    re = resample_path(ori, step=0.25, include_original=True)

    a, b = Polygon(ori), Polygon(re)
    assert np.isclose(a.area, b.area)
    assert np.isclose(a.length, b.length)


def test_resample_original_case():
    # check a case that was failing to insert the original points

    from scipy.spatial import KDTree
    from shapely.geometry import LineString

    from trimesh.path.traversal import PathSample

    test_path = np.array([[0, 0], [1, 0], [1.1, 0], [2, 0], [3, 0]], dtype=float)

    ps = PathSample(test_path)
    # note the length of the original path is 3, so this is sampling past the end
    a = ps.sample(np.arange(0, 5, 1.5), include_original=True)

    # get the distance to the closest point from our test path to the sample
    radius, index = KDTree(test_path).query(a)

    # assert sampling is increasing in index
    assert ((index[1:] - index[:-1]) >= 0).all()

    # the sample path includes points that didn't exist in the original
    # so we need to check that every index has at least one zero-radius
    radius_ok = np.zeros(len(test_path), dtype=bool)
    radius_ok[index[radius < 1e-12]] = True
    assert radius_ok.all(), "not every point was inserted!"

    # run a few more quick checks to make sure the path wasn't trashed
    a = LineString(test_path)
    b = LineString(a)

    assert np.isclose(a.length, b.length)
    assert np.allclose(a.bounds, b.bounds)


if __name__ == "__main__":
    test_resample_original()
    test_resample_original_case()



try:
    from . import generic as g
except BaseException:
    import generic as g


class VerticesTest(g.unittest.TestCase):
    def test_vertex_faces(self):
        # One watertight, one not; also various sizes
        meshes = [g.get_mesh("featuretype.STL"), g.get_mesh("cycloidal.ply")]

        for m in meshes:
            # make sure every
            vertex_faces = m.vertex_faces
            for i, v in enumerate(vertex_faces):
                # filter out negative indices
                faces = m.faces[v[v >= 0]]
                assert all(i in face for face in faces)

            # choose some random vertices and make sure their
            # face indices are correct
            rand_vertices = g.np.random.randint(low=0, high=len(m.vertices), size=100)
            for v in rand_vertices:
                v_faces = g.np.where(m.faces == v)[0][::-1]
                assert g.np.all(v_faces == m.vertex_faces[v][m.vertex_faces[v] >= 0])

            # make mesh degenerate
            m.faces[0] = [0, 0, 0]
            m.faces[1] = [1, 1, 0]
            # make sure every
            vertex_faces = m.vertex_faces
            for i, v in enumerate(vertex_faces):
                # filter out negative indices
                faces = m.faces[v[v >= 0]]
                assert all(i in face for face in faces)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class SubDivideTest(g.unittest.TestCase):
    def test_subdivide(self):
        meshes = [
            g.get_mesh("soup.stl"),  # a soup of random triangles
            g.get_mesh("cycloidal.ply"),  # a mesh with multiple bodies
            g.get_mesh("featuretype.STL"),
        ]  # a mesh with a single body

        for m in meshes:
            sub = m.subdivide()
            assert g.np.allclose(m.area, sub.area)
            assert len(sub.faces) > len(m.faces)

            max_edge = m.scale / 50
            sub, idx = m.subdivide_to_size(max_edge=max_edge, return_index=True)
            assert g.np.allclose(m.area, sub.area)
            edge_len = (
                g.np.diff(sub.vertices[sub.edges_unique], axis=1).reshape((-1, 3)) ** 2
            ).sum(axis=1) ** 0.5
            assert (edge_len < max_edge).all()

            # should be the same order of magnitude size
            assert g.np.allclose(m.extents, sub.extents, rtol=2)
            # should be one index per new face
            assert len(idx) == len(sub.faces)
            # every face should be subdivided
            assert idx.max() == (len(m.faces) - 1)

            # check the original face index using barycentric coordinates
            epsilon = 1e-3
            for vid in sub.faces.T:
                # find the barycentric coordinates
                bary = g.trimesh.triangles.points_to_barycentric(
                    m.triangles[idx], sub.vertices[vid]
                )
                # if face indexes are correct they will be on the triangle
                # which means all barycentric coordinates are between 0.0-1.0
                assert bary.max() < (1 + epsilon)
                assert bary.min() > -epsilon
                # make sure it's not all zeros
                assert g.np.ptp(bary) > epsilon

            v, f = g.trimesh.remesh.subdivide(vertices=m.vertices, faces=m.faces)

            max_edge = m.scale / 50
            v, f, idx = g.trimesh.remesh.subdivide_to_size(
                vertices=m.vertices, faces=m.faces, max_edge=max_edge, return_index=True
            )
            ms = g.trimesh.Trimesh(vertices=v, faces=f)
            assert g.np.allclose(m.area, ms.area)
            edge_len = (
                g.np.diff(ms.vertices[ms.edges_unique], axis=1).reshape((-1, 3)) ** 2
            ).sum(axis=1) ** 0.5
            assert (edge_len < max_edge).all()

            # should be one index per new face
            assert len(idx) == len(f)
            # every face should be subdivided
            assert idx.max() == (len(m.faces) - 1)

            # check the original face index using barycentric coordinates
            epsilon = 1e-3
            for vid in f.T:
                # find the barycentric coordinates
                bary = g.trimesh.triangles.points_to_barycentric(m.triangles[idx], v[vid])
                # if face indexes are correct they will be on the triangle
                # which means all barycentric coordinates are between 0.0-1.0
                assert bary.max() < (1 + epsilon)
                assert bary.min() > -epsilon
                # make sure it's not all zeros
                assert g.np.ptp(bary) > epsilon

            check = m.subdivide_to_size(
                max_edge=m.extents.sum(), max_iter=1, return_index=False
            )
            assert check.faces.shape == m.faces.shape

    def test_sub(self):
        # try on some primitives
        meshes = [g.trimesh.creation.box(), g.trimesh.creation.icosphere()]

        for m in meshes:
            # set vertex positions as attributes for trivial check after subdivision
            # make sure we're copying the array to avoid in-place check
            m.vertex_attributes = {"pos": g.np.array(m.vertices) + 1.0}

            s = m.subdivide(face_index=[0, len(m.faces) - 1])
            # shouldn't have subdivided in-place
            assert len(s.faces) > len(m.faces)
            # area should be the same
            assert g.np.isclose(m.area, s.area)
            # volume should be the same
            assert g.np.isclose(m.volume, s.volume)
            # position attributes and actual vertices should be the same
            assert g.np.allclose(s.vertex_attributes["pos"], s.vertices + 1.0)

            max_edge = m.scale / 50
            s = m.subdivide_to_size(max_edge=max_edge)
            # shouldn't have subdivided in-place
            assert len(s.faces) > len(m.faces)
            # area should be the same
            assert g.np.isclose(m.area, s.area)
            # volume should be the same
            assert g.np.isclose(m.volume, s.volume)

    def test_loop(self):
        meshes = [
            g.get_mesh("soup.stl"),  # a soup of random triangles
            g.get_mesh("featuretype.STL"),
        ]  # a mesh with a single body

        for m in meshes:
            sub = m.subdivide_loop(iterations=1)
            # number of faces should increase
            assert len(sub.faces) > len(m.faces)
            # subdivided faces are smaller
            assert sub.area_faces.mean() < m.area_faces.mean()

    def test_loop_multibody(self):
        # a mesh with multiple bodies
        mesh = g.get_mesh("cycloidal.ply")
        sub = mesh.subdivide_loop(iterations=2)

        # number of faces should increase
        assert len(sub.faces) > len(mesh.faces)
        # subdivided faces are smaller
        assert sub.area_faces.mean() < mesh.area_faces.mean()
        # should be the same order of magnitude area
        # rtol=2 means it can be up to twice/half
        assert g.np.isclose(sub.area, mesh.area, rtol=2)
        # should have the same number of bodies
        assert len(mesh.split()) == len(sub.split())

    def test_loop_multi_simple(self, count=10):
        meshes = []
        for i in range(count):
            current = g.trimesh.creation.icosahedron()
            current.apply_translation([i * 1.5, 0, 0])
            meshes.append(current)
        # concatenate into a single multibody mesh
        m = g.trimesh.util.concatenate(meshes)
        # run subdivision on that
        a = m.subdivide_loop(iterations=4)
        # make sure it splits and is watertight
        split = a.split()
        assert len(split) == count
        assert all(i.is_watertight for i in split)

    def test_loop_correct(self):
        box = g.trimesh.creation.box()
        big_sphere = g.trimesh.creation.icosphere(radius=0.5)
        small_sphere = g.trimesh.creation.icosphere(radius=0.4)
        sub = box.subdivide_loop(iterations=2)
        # smaller than 0.5 sphere
        assert big_sphere.contains(sub.vertices).all()
        # bigger than 0.4 sphere
        assert (~small_sphere.contains(sub.vertices)).all()

    def test_loop_bound(self):
        def _get_boundary_vertices(mesh):
            boundary_groups = g.trimesh.grouping.group_rows(
                mesh.edges_sorted, require_count=1
            )
            return mesh.vertices[g.np.unique(mesh.edges_sorted[boundary_groups])]

        box = g.trimesh.creation.box()
        bottom_mask = g.np.zeros(len(box.faces), dtype=bool)
        bottom_faces = [1, 5]
        bottom_mask[bottom_faces] = True
        # eliminate bottom of the box
        box.update_faces(~bottom_mask)
        bottom_vrts = _get_boundary_vertices(box)
        # subdivide box
        sub = box.subdivide_loop(iterations=2)
        sub_bottom_vrts = _get_boundary_vertices(sub)

        # y value of bottom boundary vertices should not be changed
        assert g.np.isclose(
            bottom_vrts[:, 1].mean(), sub_bottom_vrts[:, 1].mean(), atol=1e-5
        )

    def test_uv(self):
        # get a mesh with texture
        m = g.get_mesh("fuze.obj")
        # m.show()
        # get the shape of the initial mesh
        shape = m.vertices.shape
        # subdivide the mesh
        s = m.subdivide()

        # shouldn't have changed source mesh
        assert m.vertices.shape == shape
        # subdivided mesh should have more vertices
        assert s.vertices.shape[0] > shape[0]
        # should have UV coordinates matching vertices
        assert s.vertices.shape[0] == s.visual.uv.shape[0]

        # original UV coordinates with faces
        ov = m.visual.uv[m.faces]
        # subdivided mesh faces
        sv = s.visual.uv[s.faces]
        # both subdivided and original should have faces
        # that don't vary wildly
        assert g.np.ptp(ov, axis=1).mean(axis=0).max() < 0.1
        assert g.np.ptp(sv, axis=1).mean(axis=0).max() < 0.1

        max_edge = m.scale / 50
        s = m.subdivide_to_size(max_edge=max_edge)

        # shouldn't have changed source mesh
        assert m.vertices.shape == shape
        # subdivided mesh should have more vertices
        assert s.vertices.shape[0] > shape[0]
        # should have UV coordinates matching vertices
        assert s.vertices.shape[0] == s.visual.uv.shape[0]

        # original UV coordinates with faces
        ov = m.visual.uv[m.faces]
        # subdivided mesh faces
        sv = s.visual.uv[s.faces]
        # both subdivided and original should have faces
        # that don't vary wildly
        assert g.np.ptp(ov, axis=1).mean(axis=0).max() < 0.1
        assert g.np.ptp(sv, axis=1).mean(axis=0).max() < 0.1

    def test_max_iter(self):
        m = g.trimesh.creation.box()
        try:
            m.subdivide_to_size(0.01, max_iter=1)
            raise BaseException("that shouldn't have worked!")
        except ValueError:
            # this shouldn't have worked
            pass
        # this should be enough iterations
        r = m.subdivide_to_size(0.01, max_iter=10)
        assert r.is_watertight

    def test_idx_simple(self):
        vertices = g.np.array([0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0]).reshape((-1, 3)) * 10
        faces = g.np.array(
            [
                0,
                1,
                2,
                0,
                2,
                3,
            ]
        ).reshape((-1, 3))

        def test(fidx):
            v, f, idx = g.trimesh.remesh.subdivide(
                vertices, faces, face_index=fidx, return_index=True
            )
            eps = 1e-8
            for fid in fidx:
                # get the new triangles, as indicated by the index
                tri_new = v[f[idx[fid]]]

                # this is the original triangle
                original = vertices[faces[g.np.tile(fid, len(tri_new) * 3)]]

                bary = g.trimesh.triangles.points_to_barycentric(
                    triangles=original, points=tri_new.reshape((-1, 3))
                )

                assert (bary < 1 + eps).all()
                assert (bary > -eps).all()

        test([0, 1])
        test([1, 0])


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import os
import re
from collections import defaultdict, deque

import numpy as np

try:
    # `pip install pillow`
    # optional: used for textured meshes
    from PIL import Image
except BaseException as E:
    # if someone tries to use Image re-raise
    # the import error so they can debug easily
    from ..exceptions import ExceptionWrapper

    Image = ExceptionWrapper(E)

from .. import util
from ..constants import log, tol
from ..resolvers import ResolverLike
from ..typed import Dict, Loadable, Optional
from ..visual.color import to_float
from ..visual.material import SimpleMaterial
from ..visual.texture import TextureVisuals, unmerge_faces


def load_obj(
    file_obj: Loadable,
    resolver: Optional[ResolverLike] = None,
    group_material: bool = True,
    skip_materials: bool = False,
    maintain_order: bool = False,
    metadata: Optional[Dict] = None,
    split_objects: bool = False,
    split_groups: bool = False,
    **kwargs,
) -> Dict:
    """
    Load a Wavefront OBJ file into kwargs for a trimesh.Scene
    object.

    Parameters
    --------------
    file_obj : file like object
      Contains OBJ data
    resolver : trimesh.visual.resolvers.Resolver
      Allow assets such as referenced textures and
      material files to be loaded
    group_material : bool
      Group faces that share the same material
      into the same mesh.
    skip_materials
      Don't load any materials.
    maintain_order
      Make the strongest attempt possible to not reorder faces
      or vertices which may result in visual artifacts and other
      odd behavior. The OBJ data structure is quite different than
      the "flat matching array" used by Trimesh and GLTF so this may
      not be completely possible.
    split_objects
      Whenever the loader encounters an `o` directive in the OBJ
      file, split the loaded result into a new Trimesh object.
    split_groups
        Whenever the loader encounters a `g` directive in the OBJ
        file, split the loaded result into a new Trimesh object.

    Returns
    -------------
    kwargs : dict
      Keyword arguments which can be loaded by
      trimesh.exchange.load.load_kwargs into a trimesh.Scene
    """
    # get text as bytes or string blob
    text = file_obj.read()
    # if text was bytes decode into string
    text = util.decode_text(text)

    # add leading and trailing newlines so we can use the
    # same logic even if they jump directly in to data lines
    text = "\n{}\n".format(text.strip().replace("\r\n", "\n"))

    # remove backslash continuation characters and merge them into the same
    # line
    text = text.replace("\\\n", "")

    # Load Materials
    materials = {}
    mtl_position = text.find("mtllib")
    if not skip_materials and mtl_position >= 0:
        # take the line of the material file after `mtllib`
        # which should be the file location of the .mtl file
        mtl_path = text[mtl_position + 6 : text.find("\n", mtl_position)].strip()
        try:
            # use the resolver to get the data
            material_kwargs = parse_mtl(resolver[mtl_path], resolver=resolver)
            # turn parsed kwargs into material objects
            materials = {k: SimpleMaterial(**v) for k, v in material_kwargs.items()}
        except (OSError, TypeError):
            # usually the resolver couldn't find the asset
            log.debug(f"unable to load materials from: {mtl_path}")
        except BaseException:
            # something else happened so log a warning
            log.debug(f"unable to load materials from: {mtl_path}", exc_info=True)

    # extract vertices from raw text
    v, vn, vt, vc = _parse_vertices(text=text)

    # get relevant chunks that have face data
    # in the form of (material, object, chunk)
    face_tuples = _preprocess_faces(text, split_objects, split_groups)

    # combine chunks that have the same material
    # some meshes end up with a LOT of components
    # and will be much slower if you don't do this
    if group_material or split_objects or split_groups:
        face_tuples = _group_by(face_tuples, group_material, split_objects, split_groups)

    # no faces but points given
    # return point cloud
    if not len(face_tuples) and v is not None:
        pc = {"vertices": v}
        if vn is not None:
            pc["vertex_normals"] = vn
        if vc is not None:
            pc["vertex_colors"] = vc
        return pc

    # Load Faces
    # now we have clean- ish faces grouped by material and object
    # so now we have to turn them into numpy arrays and kwargs
    # for trimesh mesh and scene objects
    geometry = {}
    while len(face_tuples) > 0:
        # consume the next chunk of text
        material, current_object, current_group, chunk = face_tuples.pop()
        # do wangling in string form
        # we need to only take the face line before a newline
        # using builtin functions in a list comprehension
        # is pretty fast relative to other options
        # this operation is the only one that is O(len(faces))
        # slower due to the tight-loop conditional:
        # face_lines = [i[:i.find('\n')]
        #              for i in chunk.split('\nf ')[1:]
        #              if i.rfind('\n') >0]
        # maxsplit=1 means that it can stop working
        # after it finds the first newline
        # passed as arg as it's not a kwarg in python2
        face_lines = [
            i.split("\n", 1)[0].strip()
            for i in re.split("^f", chunk, flags=re.MULTILINE)[1:]
        ]

        # check every face for mixed tri-quad-ngon
        columns = len(face_lines[0].replace("/", " ").split())
        flat_array = all(columns == len(f.replace("/", " ").split()) for f in face_lines)

        # make sure we have the right number of values for vectorized
        if flat_array:
            # the fastest way to get to a numpy array
            # processes the whole string at once into a 1D array
            array = np.fromstring(
                " ".join(face_lines).replace("/", " "), sep=" ", dtype=np.int64
            )
            # also wavefront is 1-indexed (vs 0-indexed) so offset
            # only applies to positive indices
            array[array > 0] -= 1

            # everything is a nice 2D array
            faces, faces_tex, faces_norm = _parse_faces_vectorized(
                array=array, columns=columns, sample_line=face_lines[0]
            )
        else:
            # if we had something annoying like mixed in quads
            # or faces that differ per-line we have to loop
            # i.e. something like:
            #  '31407 31406 31408',
            #  '32303/2469 32304/2469 32305/2469',
            log.debug("faces have mixed data: using slow fallback!")
            faces, faces_tex, faces_norm = _parse_faces_fallback(face_lines)

        # build name from components
        name_parts = []
        if split_objects and current_object is not None:
            name_parts.append(current_object)
        if split_groups and current_group is not None:
            name_parts.append(current_group)
        if group_material and len(materials) > 1 and material is not None:
            name_parts.append(str(material))

        # join parts or fall back to defaults
        if name_parts:
            name = "_".join(name_parts)
        elif current_object is not None:
            name = current_object
        else:
            # try to use the file name from the resolver
            # or file object if possible before defaulting
            name = next(
                i
                for i in (
                    getattr(resolver, "file_name", None),
                    getattr(file_obj, "name", None),
                    "geometry",
                )
                if i is not None
            )

        # ensure the name is always unique in the geometry dict
        name = util.unique_name(name, geometry)

        # try to get usable texture
        mesh = kwargs.copy()
        if faces_tex is not None:
            # convert faces referencing vertices and
            # faces referencing vertex texture to new faces
            # where each face
            if faces_norm is not None and len(faces_norm) == len(faces):
                new_faces, mask_v, mask_vt, mask_vn = unmerge_faces(
                    faces, faces_tex, faces_norm, maintain_faces=maintain_order
                )
            else:
                mask_vn = None
                # no face normals but face texturre
                new_faces, mask_v, mask_vt = unmerge_faces(
                    faces, faces_tex, maintain_faces=maintain_order
                )

            if tol.strict:
                # we should NOT have messed up the faces
                # note: this is EXTREMELY slow due to all the
                # float comparisons so only run this in unit tests
                assert np.allclose(v[faces], v[mask_v][new_faces])
                # faces should all be in bounds of vertives
                assert new_faces.max() < len(v[mask_v])
            try:
                # survive index errors as sometimes we
                # want materials without UV coordinates
                uv = vt[mask_vt]
            except BaseException:
                log.debug("index failed on UV coordinates, skipping!")
                uv = None

            # mask vertices and use new faces
            mesh.update({"vertices": v[mask_v].copy(), "faces": new_faces})

        else:
            # otherwise just use unmasked vertices
            uv = None
            # check to make sure indexes are in bounds
            if tol.strict:
                assert faces.max() < len(v)
            if vn is not None and np.shape(faces_norm) == faces.shape:
                # do the crazy unmerging logic for split indices
                new_faces, mask_v, mask_vn = unmerge_faces(
                    faces, faces_norm, maintain_faces=maintain_order
                )
            else:
                # face_tex is None and
                # generate the mask so we only include
                # referenced vertices in every new mesh
                if maintain_order:
                    mask_v = np.ones(len(v), dtype=bool)
                else:
                    mask_v = np.zeros(len(v), dtype=bool)
                mask_v[faces] = True

                # reconstruct the faces with the new vertex indices
                inverse = np.zeros(len(v), dtype=np.int64)
                inverse[mask_v] = np.arange(mask_v.sum())
                new_faces = inverse[faces]
                # no normals
                mask_vn = None

            # start with vertices and faces
            mesh.update({"faces": new_faces, "vertices": v[mask_v].copy()})

        # if colors and normals are OK save them
        if vc is not None:
            try:
                # may fail on a malformed color mask
                mesh["vertex_colors"] = vc[mask_v]
            except BaseException:
                log.debug("failed to load vertex_colors", exc_info=True)
        if mask_vn is not None:
            try:
                # may fail on a malformed mask
                normals = vn[mask_vn]
                if normals.shape != mesh["vertices"].shape:
                    raise ValueError(
                        "incorrect normals {} != {}".format(
                            str(normals.shape), str(mesh["vertices"].shape)
                        )
                    )
                mesh["vertex_normals"] = normals
            except BaseException:
                log.debug("failed to load vertex_normals", exc_info=True)

        visual = None
        if material in materials:
            # use the material with the UV coordinates
            visual = TextureVisuals(uv=uv, material=materials[material])
        elif uv is not None and len(uv) == len(mesh["vertices"]):
            # create a texture with an empty materials
            visual = TextureVisuals(uv=uv)
        elif material is not None:
            # case where material is specified but not available
            log.debug(f"specified material ({material})  not loaded!")
        # assign the visual
        mesh["visual"] = visual
        # store geometry by name
        geometry[name] = mesh

    # add an identity transform for every geometry
    graph = [{"geometry": k, "frame_to": k} for k in geometry.keys()]

    # convert to scene kwargs
    return {"geometry": geometry, "graph": graph}


def parse_mtl(mtl, resolver=None):
    """
    Parse a loaded MTL file.

    Parameters
    -------------
    mtl : str or bytes
      Data from an MTL file
    resolver : trimesh.Resolver
      Fetch assets by name from file system, web, or other

    Returns
    ------------
    mtllibs : list of dict
      Each dict has keys: newmtl, map_Kd, Kd
    """
    # decode bytes into string if necessary
    mtl = util.decode_text(mtl)

    # current material
    material = None
    # materials referenced by name
    materials = {}
    # use universal newline splitting
    lines = str.splitlines(str(mtl).strip())

    # remap OBJ property names to kwargs for SimpleMaterial
    mapped = {"kd": "diffuse", "ka": "ambient", "ks": "specular", "ns": "glossiness"}

    for line in lines:
        # split by white space
        split = line.strip().split()
        # needs to be at least two values
        if len(split) <= 1:
            continue
        # the first value is the parameter name
        key = split[0].lower()
        # start a new material
        if key == "newmtl":
            # material name extracted from line like:
            # newmtl material_0
            if material is not None:
                # save the old material by old name and remove key
                materials[material["name"]] = material
            # start a fresh new material
            # do we really want to support material names with whitespace?
            material = {"name": " ".join(split[1:])}

        elif key == "map_kd":
            # represents the file name of the texture image
            index = line.lower().index("map_kd") + 6
            file_name = line[index:].strip()
            try:
                file_data = resolver.get(file_name)
                # load the bytes into a PIL image
                # an image file name
                material["image"] = Image.open(util.wrap_as_stream(file_data))
                # also store the original map_kd file name
                material["image"].info["file_path"] = os.path.abspath(
                    os.path.join(getattr(resolver, "parent", ""), file_name)
                )

            except BaseException:
                log.debug("failed to load image", exc_info=True)

        elif key in mapped.keys():
            try:
                # diffuse, ambient, and specular float RGB
                value = [float(x) for x in split[1:]]
                # if there is only one value return that
                if len(value) == 1:
                    value = value[0]
                if material is not None:
                    # store the key by mapped value
                    material[mapped[key]] = value
                    # also store key by OBJ name
                    material[key] = value
            except BaseException:
                log.debug("failed to convert color!", exc_info=True)
        # pass everything as kwargs to material constructor
        elif material is not None:
            # save any other unspecified keys
            material[key] = split[1:]
    # reached EOF so save any existing materials
    if material:
        materials[material["name"]] = material

    return materials


def _parse_faces_vectorized(array, columns, sample_line):
    """
    Parse loaded homogeneous (tri/quad) face data in a
    vectorized manner.

    Parameters
    ------------
    array : (n,) int
      Indices in order
    columns : int
      Number of columns in the file
    sample_line : str
      A single line so we can assess the ordering

    Returns
    --------------
    faces : (n, d) int
      Faces in space
    faces_tex : (n, d) int or None
      Texture for each vertex in face
    faces_norm : (n, d) int or None
      Normal index for each vertex in face
    """
    # reshape to columns
    array = array.reshape((-1, columns))
    # how many elements are in the first line of faces
    # i.e '13/1/13 14/1/14 2/1/2 1/2/1' is 4
    group_count = len(sample_line.strip().split())
    # how many elements are there for each vertex reference
    # i.e. '12/1/13' is 3
    per_ref = int(columns / group_count)
    # create an index mask we can use to slice vertex references
    index = np.arange(group_count) * per_ref
    # slice the faces out of the blob array
    faces = array[:, index]

    # TODO: probably need to support 8 and 12 columns for quads
    # or do something more general
    faces_tex, faces_norm = None, None
    if columns == group_count * 2:
        # if we have two values per vertex the second
        # one is index of texture coordinate (`vt`)
        # count how many delimiters are in the first face line
        # to see if our second value is texture or normals
        # do splitting to clip off leading/trailing slashes
        count = "".join(i.strip("/") for i in sample_line.split()).count("/")
        if count == columns:
            # case where each face line looks like:
            # ' 75//139 76//141 77//141'
            # which is vertex/nothing/normal
            faces_norm = array[:, index + 1]
        elif count == int(columns / 2):
            # case where each face line looks like:
            # '75/139 76/141 77/141'
            # which is vertex/texture
            faces_tex = array[:, index + 1]
        else:
            log.debug(f"face lines are weird: {sample_line}")
    elif columns == group_count * 3:
        # if we have three values per vertex
        # second value is always texture
        faces_tex = array[:, index + 1]
        # third value is reference to vertex normal (`vn`)
        faces_norm = array[:, index + 2]
    return faces, faces_tex, faces_norm


def _parse_faces_fallback(lines):
    """
    Use a slow but more flexible looping method to process
    face lines as a fallback option to faster vectorized methods.

    Parameters
    -------------
    lines : (n,) str
      List of lines with face information

    Returns
    -------------
    faces : (m, 3) int
      Clean numpy array of face triangles
    """

    # collect vertex, texture, and vertex normal indexes
    v, vt, vn = [], [], []

    # loop through every line starting with a face
    for line in lines:
        # remove leading newlines then
        # take first bit before newline then split by whitespace
        split = line.strip().split("\n")[0].split()
        # split into: ['76/558/76', '498/265/498', '456/267/456']
        len_split = len(split)
        if len_split == 3:
            pass
        elif len_split == 4:
            # triangulate quad face
            split = [split[0], split[1], split[2], split[2], split[3], split[0]]
        elif len_split > 4:
            # triangulate polygon as a triangles fan
            collect = []
            # we need a flat list so append inside
            # a list comprehension
            collect_append = collect.append
            [
                [
                    collect_append(split[0]),
                    collect_append(split[i + 1]),
                    collect_append(split[i + 2]),
                ]
                for i in range(len(split) - 2)
            ]
            split = collect
        else:
            log.debug(f"face needs more values 3>{len(split)} skipping!")
            continue

        # f is like: '76/558/76'
        for f in split:
            # vertex, vertex texture, vertex normal
            split = f.split("/")
            # we always have a vertex reference
            v.append(int(split[0]))

            # faster to try/except than check in loop
            try:
                vt.append(int(split[1]))
            except BaseException:
                pass
            try:
                # vertex normal is the third index
                vn.append(int(split[2]))
            except BaseException:
                pass

    # shape into triangles and switch to 0-indexed
    # 0-indexing only applies to positive indices
    faces = np.array(v, dtype=np.int64).reshape((-1, 3))
    faces[faces > 0] -= 1
    faces_tex, normals = None, None
    if len(vt) == len(v):
        faces_tex = np.array(vt, dtype=np.int64).reshape((-1, 3))
        faces_tex[faces_tex > 0] -= 1
    if len(vn) == len(v):
        normals = np.array(vn, dtype=np.int64).reshape((-1, 3))
        normals[normals > 0] -= 1

    return faces, faces_tex, normals


def _parse_vertices(text):
    """
    Parse raw OBJ text into vertices, vertex normals,
    vertex colors, and vertex textures.

    Parameters
    -------------
    text : str
      Full text of an OBJ file

    Returns
    -------------
    v : (n, 3) float
      Vertices in space
    vn : (m, 3) float or None
      Vertex normals
    vt : (p, 2) float or None
      Vertex texture coordinates
    vc : (n, 3) float or None
      Per-vertex color
    """

    # the first position of a vertex in the text blob
    # we only really need to search from the start of the file
    # up to the location of out our first vertex but we
    # are going to use this check for "do we have texture"
    # determination later so search the whole stupid file
    starts = {k: text.find(f"\n{k} ") for k in ["v", "vt", "vn"]}

    # no valid values so exit early
    if not any(v >= 0 for v in starts.values()):
        return None, None, None, None

    # find the last position of each valid value
    ends = {
        k: text.find("\n", text.rfind(f"\n{k} ") + 2 + len(k))
        for k, v in starts.items()
        if v >= 0
    }

    # take the first and last position of any vertex property
    start = min(s for s in starts.values() if s >= 0)
    end = max(e for e in ends.values() if e >= 0)
    # get the chunk of test that contains vertex data
    chunk = text[start:end].replace("+e", "e").replace("-e", "e")

    # get the clean-ish data from the file as python lists
    data = {
        k: [i.split("\n", 1)[0] for i in chunk.split(f"\n{k} ")[1:]]
        for k, v in starts.items()
        if v >= 0
    }

    # count the number of data values per row on a sample row
    per_row = {k: len(v[0].split()) for k, v in data.items()}

    # convert data values into numpy arrays
    result = defaultdict(lambda: None)
    for k, value in data.items():
        # use joining and fromstring to get as numpy array
        array = np.fromstring(" ".join(value), sep=" ", dtype=np.float64)
        # what should our shape be
        shape = (len(value), per_row[k])
        # check shape of flat data
        if len(array) == np.prod(shape):
            # we have a nice 2D array
            result[k] = array.reshape(shape)
        else:
            # we don't have a nice (n, d) array so fall back to a slow loop
            # this is where mixed "some of the values but not all have vertex colors"
            # problem is handled.
            lines = []
            [[lines.append(v.strip().split()) for v in str.splitlines(i)] for i in value]
            # we need to make a 2D array so clip it to the shortest array
            count = min(len(L) for L in lines)
            # make a numpy array out of the cleaned up line data
            result[k] = np.array([L[:count] for L in lines], dtype=np.float64)

    # vertices
    v = result["v"]
    # vertex colors are stored next to vertices
    vc = None
    if v is not None and v.shape[1] >= 6:
        # vertex colors are stored after vertices
        v, vc = v[:, :3], v[:, 3:6]
    elif v is not None and v.shape[1] > 3:
        # we got a lot of something unknowable
        v = v[:, :3]

    # vertex texture or None
    vt = result["vt"]
    if vt is not None:
        # sometimes UV coordinates come in as UVW
        vt = vt[:, :2]
    # vertex normals or None
    vn = result["vn"]

    # check will generally only be run in unit tests
    # so we are allowed to do things that are slow
    if tol.strict:
        # check to make sure our subsetting
        # didn't miss any vertices or data
        assert len(v) == text.count("\nv ")
        # make sure optional data matches file too
        if vn is not None:
            assert len(vn) == text.count("\nvn ")
        if vt is not None:
            assert len(vt) == text.count("\nvt ")

    return v, vn, vt, vc


def _group_by(face_tuples, use_mtl: bool, use_obj: bool, use_group: bool):
    """
    For chunks of faces split by material group
    the chunks that share the same material.

    Parameters
    ------------
    face_tuples : (n,) list of (material, obj, chunk)
      The data containing faces
    use_mtl
      Group tuples by `usemtl` commands
    use_obj
      Group tuples by `o` commands
    use_group
      Group tuples by `g` commands

    Returns
    ------------
    grouped : (m,) list of tuples containing:
              `(material name, objectname, group name, raw chunk)`
    """

    # store the chunks grouped by material
    grouped = defaultdict(lambda: ["", "", "", []])
    # loop through existring
    for material, obj, group, chunk in face_tuples:
        # tuple key for the dict
        key = (
            material if use_mtl else None,
            obj if use_obj else None,
            group if use_group else None,
        )
        grouped[key][0] = material
        grouped[key][1] = obj
        grouped[key][2] = group
        # don't do a million string concatenations in loop
        grouped[key][3].append(chunk)
    # go back and do a join to make a single string
    for key in grouped.keys():
        grouped[key][3] = "\n".join(grouped[key][3])
    # return as list
    return list(grouped.values())


def _preprocess_faces(text, use_obj=False, use_groups=False):
    """
    Pre-Process Face Text

    Rather than looking at each line in a loop we're
    going to split lines by directives which indicate
    a new mesh, specifically 'usemtl', 'o', and 'g' keys
    search for materials, objects, faces, or groups

    Parameters
    ------------
    text : str
      Raw file

    Returns
    ------------
    triple : (n, 3) tuple
      Tuples of (material, object, data-chunk)
    """
    # see which chunk is relevant
    starters = ["\nusemtl ", "\no ", "\nf ", "\ng ", "\ns "]
    f_start = len(text)
    # first index of material, object, face, group, or smoother
    for st in starters:
        search = text.find(st, 0, f_start)
        # if not contained find will return -1
        if search < 0:
            continue
        # subtract the length of the key from the position
        # to make sure it's included in the slice of text
        if search < f_start:
            f_start = search
    # index in blob of the newline after the last face
    f_end = text.find("\n", text.rfind("\nf ") + 3)
    # get the chunk of the file that has face information
    if f_end >= 0:
        # clip to the newline after the last face
        f_chunk = text[f_start:f_end]
    else:
        # no newline after last face
        f_chunk = text[f_start:]

    if tol.strict:
        # check to make sure our subsetting didn't miss any faces
        assert f_chunk.count("\nf ") == text.count("\nf ")

    # two things cause new meshes to be created:
    # objects and materials
    # re.finditer was faster than find in a loop
    # find the index of every material change
    idx_mtl = np.array([m.start(0) for m in re.finditer("usemtl ", f_chunk)], dtype=int)
    # find the index of every new object
    split_idxs = [[0, len(f_chunk)], idx_mtl]

    # NOTE: This used to split objects on every run, but now it does not
    if use_obj:
        split_idxs.append(
            np.array([m.start(0) for m in re.finditer("\no ", f_chunk)], dtype=int)
        )

    if use_groups:
        split_idxs.append(
            np.array([m.start(0) for m in re.finditer("\ng ", f_chunk)], dtype=int)
        )

    # find all the indexes where we want to split
    splits = np.unique(np.concatenate(tuple(split_idxs)))

    # track the current material and object ID
    current_obj = None
    current_mtl = None
    current_group = None
    # store (material, object, group, face lines)
    face_tuples = []

    for start, end in zip(splits[:-1], splits[1:]):
        # ensure there's always a trailing newline
        chunk = f_chunk[start:end].strip() + "\n"
        if chunk.startswith("o "):
            current_obj, chunk = chunk.split("\n", 1)
            current_obj = current_obj[2:].strip()
        elif chunk.startswith("usemtl"):
            current_mtl, chunk = chunk.split("\n", 1)
            current_mtl = current_mtl[6:].strip()
        # Discard the g tag line in the list of faces
        elif chunk.startswith("g "):
            current_group, chunk = chunk.split("\n", 1)
            current_group = current_group[2:].strip()
        # If we have an f at the beginning of a line
        # then add it to the list of faces chunks
        if chunk.startswith("f ") or "\nf" in chunk:
            face_tuples.append((current_mtl, current_obj, current_group, chunk))
    return face_tuples


def export_obj(
    mesh,
    include_normals=None,
    include_color=True,
    include_texture=True,
    return_texture=False,
    write_texture=True,
    resolver=None,
    digits=8,
    mtl_name=None,
    header="https://github.com/mikedh/trimesh",
):
    """
    Export a mesh as a Wavefront OBJ file.
    TODO: scenes with textured meshes

    Parameters
    -----------
    mesh : trimesh.Trimesh
      Mesh to be exported
    include_normals : Optional[bool]
      Include vertex normals in export. If None
      will only be included if vertex normals are in cache.
    include_color : bool
      Include vertex color in export
    include_texture : bool
      Include `vt` texture in file text
    return_texture : bool
      If True, return a dict with texture files
    write_texture : bool
      If True and a writable resolver is passed
      write the referenced texture files with resolver
    resolver : None or trimesh.resolvers.Resolver
      Resolver which can write referenced text objects
    digits : int
      Number of digits to include for floating point
    mtl_name : None or str
      If passed, the file name of the MTL file.
    header : str or None
      Header string for top of file or None for no header.

    Returns
    -----------
    export : str
      OBJ format output
    texture : dict
      Contains files that need to be saved in the same
      directory as the exported mesh: {file name : bytes}
    """
    # store the multiple options for formatting
    # vertex indexes for faces
    face_formats = {
        ("v",): "{}",
        ("v", "vn"): "{}//{}",
        ("v", "vt"): "{}/{}",
        ("v", "vn", "vt"): "{}/{}/{}",
    }

    # check the input
    if util.is_instance_named(mesh, "Trimesh"):
        meshes = [mesh]
    elif util.is_instance_named(mesh, "Scene"):
        meshes = mesh.dump()
    elif util.is_instance_named(mesh, "PointCloud"):
        meshes = [mesh]
    else:
        raise ValueError("must be Trimesh or Scene!")

    # collect lines to export
    objects = deque([])
    # keep track of the number of each export element
    counts = {"v": 0, "vn": 0, "vt": 0}
    # collect materials as we go
    materials = {}
    materials_name = set()

    for current in meshes:
        # we are going to reference face_formats with this
        face_type = ["v"]
        # OBJ includes vertex color as RGB elements on the same line
        if (
            include_color
            and current.visual.kind in ["vertex", "face"]
            and len(current.visual.vertex_colors)
        ):
            # create a stacked blob with position and color
            v_blob = np.column_stack(
                (current.vertices, to_float(current.visual.vertex_colors[:, :3]))
            )
        else:
            # otherwise just export vertices
            v_blob = current.vertices

            # add the first vertex key and convert the array
        # add the vertices
        export = deque(
            [
                "v "
                + util.array_to_string(
                    v_blob, col_delim=" ", row_delim="\nv ", digits=digits
                )
            ]
        )

        # if include_normals is None then
        # only include if they're already stored
        if include_normals is None:
            include_normals = "vertex_normals" in current._cache.cache

        if include_normals:
            try:
                converted = util.array_to_string(
                    current.vertex_normals,
                    col_delim=" ",
                    row_delim="\nvn ",
                    digits=digits,
                )
                # if vertex normals are stored in cache export them
                face_type.append("vn")
                export.append("vn " + converted)
            except BaseException:
                log.debug("failed to convert vertex normals", exc_info=True)

        # collect materials into a dict
        if include_texture and hasattr(current.visual, "uv"):
            try:
                # get a SimpleMaterial
                material = current.visual.material
                if hasattr(material, "to_simple"):
                    material = material.to_simple()

                # hash the material to avoid duplicates
                hashed = hash(material)
                if hashed not in materials:
                    # get a unique name for the material
                    name = util.unique_name(material.name, materials_name)
                    # add the name to our collection
                    materials_name.add(name)
                    # convert material to an OBJ MTL
                    materials[hashed] = material.to_obj(name=name)

                # get the name of the current material as-stored
                tex_name = materials[hashed][1]

                # export the UV coordinates
                if len(np.shape(getattr(current.visual, "uv", None))) == 2:
                    converted = util.array_to_string(
                        current.visual.uv, col_delim=" ", row_delim="\nvt ", digits=digits
                    )
                    # if vertex texture exists and is the right shape
                    face_type.append("vt")
                    # add the uv coordinates
                    export.append("vt " + converted)
                # add the directive to use the exported material
                export.appendleft(f"usemtl {tex_name}")
            except BaseException:
                log.debug("failed to convert UV coordinates", exc_info=True)

        # the format for a single vertex reference of a face
        face_format = face_formats[tuple(face_type)]
        # add the exported faces to the export if available
        if hasattr(current, "faces"):
            export.append(
                "f "
                + util.array_to_string(
                    current.faces + 1 + counts["v"],
                    col_delim=" ",
                    row_delim="\nf ",
                    value_format=face_format,
                )
            )
        # offset our vertex position
        counts["v"] += len(current.vertices)

        # add object name if found in metadata
        if "name" in current.metadata:
            export.appendleft("\no {}".format(current.metadata["name"]))
        # add this object
        objects.append("\n".join(export))

    # collect files like images to write
    mtl_data = {}
    # combine materials
    if len(materials) > 0:
        # collect text for a single mtllib file
        mtl_lib = []
        # now loop through: keys are garbage hash
        # values are (data, name)
        for data, _ in materials.values():
            for file_name, file_data in data.items():
                if file_name.lower().endswith(".mtl"):
                    # collect mtl lines into single file
                    mtl_lib.append(file_data)
                elif file_name not in mtl_data:
                    # things like images
                    mtl_data[file_name] = file_data
                else:
                    log.warning(f"not writing {file_name}")

        if mtl_name is None:
            # if no name passed set a default
            mtl_name = "material.mtl"

        # prepend a header to the MTL text if requested
        if header is not None:
            prepend = f"# {header}\n\n".encode()
        else:
            prepend = b""

        # save the material data
        mtl_data[mtl_name] = prepend + b"\n\n".join(mtl_lib)
        # add the reference to the MTL file
        objects.appendleft(f"mtllib {mtl_name}")

    if header is not None:
        # add a created-with header to the top of the file
        objects.appendleft(f"# {header}")

    # add a trailing newline
    objects.append("\n")

    # combine elements into a single string
    text = "\n".join(objects)

    # if we have a resolver and have asked to write texture
    if write_texture and resolver is not None and len(materials) > 0:
        # not all resolvers have a write method
        [resolver.write(k, v) for k, v in mtl_data.items()]

    # if we exported texture it changes returned values
    if return_texture:
        return text, mtl_data

    return text


_obj_loaders = {"obj": load_obj}



"""
convex.py

Deal with creating and checking convex objects in 2, 3 and N dimensions.

Convex is defined as:
1) "Convex, meaning "curving out" or "extending outward" (compare to concave)
2) having an outline or surface curved like the exterior of a circle or sphere.
3) (of a polygon) having only interior angles measuring less than 180
"""

from dataclasses import dataclass, fields

import numpy as np

from . import triangles, util
from .constants import tol
from .parent import Geometry3D
from .typed import NDArray, Union

try:
    from scipy.spatial import ConvexHull
except ImportError as E:
    from .exceptions import ExceptionWrapper

    ConvexHull = ExceptionWrapper(E)

try:
    from scipy.spatial import QhullError
except BaseException:
    QhullError = BaseException


@dataclass
class QhullOptions:
    """
    A helper class for constructing correct Qhull option strings.
    More details available at: http://www.qhull.org/html/qh-quick.htm#options

    Currently only includes the boolean flag options, which is most of them.

    Parameters
    -----------
    Qa
      Allow input with fewer or more points than coordinates
    Qc
      Keep coplanar points with nearest facet
    Qi
      Keep interior points with nearest facet.
    QJ
      Joggled input to avoid precision problems
    Qt
      Triangulated output.
    Qu
      Compute upper hull for furthest-site Delaunay triangulation
    Qw
      Allow warnings about Qhull options
    Qbb
      Scale last coordinate to [0,m] for Delaunay
    Qs
      Search all points for the initial simplex
    Qv
      Test vertex neighbors for convexity
    Qx
      Exact pre-merges (allows coplanar facets)
    Qz
      Add a point-at-infinity for Delaunay triangulations
    QbB
      Scale input to fit the unit cube
    QR0
      Random rotation (n=seed, n=0 time, n=-1 time/no rotate)
    Qg
      only build good facets (needs 'QGn', 'QVn', or 'Pdk')
    Pp
      Do not print statistics about precision problems and remove
      some of the warnings including the narrow hull warning.
    """

    Qa: bool = False
    """ Allow input with fewer or more points than coordinates"""

    Qc: bool = False
    """ Keep coplanar points with nearest facet"""

    Qi: bool = False
    """ Keep interior points with nearest facet. """

    QJ: bool = False
    """ Joggled input to avoid precision problems """

    Qt: bool = False
    """ Triangulated output. """

    Qu: bool = False
    """ Compute upper hull for furthest-site Delaunay triangulation """

    Qw: bool = False
    """ Allow warnings about Qhull options """

    # Precision handling
    Qbb: bool = False
    """ Scale last coordinate to [0,m] for Delaunay """

    Qs: bool = False
    """ Search all points for the initial simplex """

    Qv: bool = False
    """ Test vertex neighbors for convexity """

    Qx: bool = False
    """ Exact pre-merges (allows coplanar facets)  """

    Qz: bool = False
    """ Add a point-at-infinity for Delaunay triangulations """

    QbB: bool = False
    """ Scale input to fit the unit cube """

    QR0: bool = False
    """ Random rotation (n=seed, n=0 time, n=-1 time/no rotate) """

    # Select facets
    Qg: bool = False
    """ Only build good facets (needs 'QGn', 'QVn', or 'Pdk') """

    Pp: bool = False
    """ Do not print statistics about precision problems and remove
      some of the warnings including the narrow hull warning. """

    # TODO : not included non-boolean options
    # QBk: Optional[Floating] = None
    # """ Scale coord[k] to upper bound of n (default 0.5) """

    # Qbk: Optional[Floating] = None
    # """ Scale coord[k] to low bound of n (default -0.5) """

    # Qbk:0Bk:0
    # """ drop dimension k from input """

    # QGn
    #    good facet if visible from point n, -n for not visible

    # QVn
    #    good facet if it includes point n, -n if not

    def __str__(self) -> str:
        """
        Construct the `qhull_options` string used by `scipy.spatial`
        objects and functions.

        Returns
        ----------
        qhull_options
          Can be passed to `scipy.spatial.[ConvexHull,Delaunay,Voronoi]`
        """
        return " ".join(f.name for f in fields(self) if getattr(self, f.name))


QHULL_DEFAULT = QhullOptions(QbB=True, Pp=True, Qt=True)


def convex_hull(
    obj: Union[Geometry3D, NDArray],
    qhull_options: Union[QhullOptions, str, None] = QHULL_DEFAULT,
    repair: bool = True,
) -> "trimesh.Trimesh":  # noqa: F821
    """
    Get a new Trimesh object representing the convex hull of the
    current mesh attempting to return a watertight mesh with correct
    normals.

    Arguments
    --------
    obj
      Mesh or `(n, 3)` points.
    qhull_options
      Options to pass to qhull.

    Returns
    --------
    convex
      Mesh of convex hull.
    """
    # would be a circular import at the module level
    from .base import Trimesh

    # compose the
    if qhull_options is None:
        qhull_str = None
    elif isinstance(qhull_options, QhullOptions):
        # use the __str__ method to compose this options string
        qhull_str = str(qhull_options)
    elif isinstance(qhull_options, str):
        qhull_str = qhull_options
    else:
        raise TypeError(type(qhull_options))

    if hasattr(obj, "vertices"):
        points = obj.vertices.view(np.ndarray)
    else:
        # will remove subclassing
        points = np.asarray(obj, dtype=np.float64)
        if not util.is_shape(points, (-1, 3)):
            raise ValueError("Object must be Trimesh or (n,3) points!")

    try:
        hull = ConvexHull(points, qhull_options=qhull_str)
    except QhullError:
        util.log.debug("Failed to compute convex hull: retrying with `QJ`", exc_info=True)
        # try with "joggle" enabled
        hull = ConvexHull(points, qhull_options="QJ")

    # hull object doesn't remove unreferenced vertices
    # create a mask to re- index faces for only referenced vertices
    vid = np.sort(hull.vertices)
    mask = np.zeros(len(hull.points), dtype=np.int64)
    mask[vid] = np.arange(len(vid))
    # remove unreferenced vertices here
    faces = mask[hull.simplices].copy()
    # rescale vertices back to original size
    vertices = hull.points[vid].copy()

    if not repair:
        # create the Trimesh object for the convex hull
        return Trimesh(vertices=vertices, faces=faces, process=True, validate=False)

    # qhull returns faces with random winding
    # calculate the returned normal of each face
    crosses = triangles.cross(vertices[faces])

    # qhull returns zero magnitude faces like an asshole
    normals, valid = util.unitize(crosses, check_valid=True)

    # remove zero magnitude faces
    faces = faces[valid]
    crosses = crosses[valid]

    # each triangle area and mean center
    triangles_area = triangles.area(crosses=crosses)
    triangles_center = vertices[faces].mean(axis=1)

    # since the convex hull is (hopefully) convex, the vector from
    # the centroid to the center of each face
    # should have a positive dot product with the normal of that face
    # if it doesn't it is probably backwards
    # note that this sometimes gets screwed up by precision issues
    centroid = np.average(triangles_center, weights=triangles_area, axis=0)
    # a vector from the centroid to a point on each face
    test_vector = triangles_center - centroid
    # check the projection against face normals
    backwards = util.diagonal_dot(normals, test_vector) < 0.0

    # flip the winding outward facing
    faces[backwards] = np.fliplr(faces[backwards])
    # flip the normal
    normals[backwards] *= -1.0

    # save the work we did to the cache so it doesn't have to be recomputed
    initial_cache = {
        "triangles_cross": crosses,
        "triangles_center": triangles_center,
        "area_faces": triangles_area,
        "centroid": centroid,
    }

    # create the Trimesh object for the convex hull
    convex = Trimesh(
        vertices=vertices,
        faces=faces,
        face_normals=normals,
        initial_cache=initial_cache,
        process=True,
        validate=False,
    )

    # we did the gross case above, but sometimes precision issues
    # leave some faces backwards anyway
    # this call will exit early if the winding is consistent
    # and if not will fix it by traversing the adjacency graph
    convex.fix_normals(multibody=False)

    # sometimes the QbB option will cause precision issues
    # so try the hull again without it and
    # check for qhull_options is None to avoid infinite recursion
    if qhull_options is None and not convex.is_winding_consistent:
        return convex_hull(convex, qhull_options=None)

    return convex


def adjacency_projections(mesh):
    """
    Test if a mesh is convex by projecting the vertices of
    a triangle onto the normal of its adjacent face.

    Parameters
    ----------
    mesh : Trimesh
      Input geometry

    Returns
    ----------
    projection : (len(mesh.face_adjacency),) float
      Distance of projection of adjacent vertex onto plane
    """
    # normals and origins from the first column of face adjacency
    normals = mesh.face_normals[mesh.face_adjacency[:, 0]]
    # one of the vertices on the shared edge
    origins = mesh.vertices[mesh.face_adjacency_edges[:, 0]]

    # faces from the second column of face adjacency
    vid_other = mesh.face_adjacency_unshared[:, 1]
    vector_other = mesh.vertices[vid_other] - origins

    # get the projection with a dot product
    dots = util.diagonal_dot(vector_other, normals)

    return dots


def is_convex(mesh):
    """
    Check if a mesh is convex.

    Parameters
    -----------
    mesh : Trimesh
      Input geometry

    Returns
    -----------
    convex : bool
      Was passed mesh convex or not
    """
    # non-watertight meshes are not convex
    # meshes with multiple bodies are not convex
    if not mesh.is_watertight or mesh.body_count != 1:
        return False

    # don't consider zero- area faces
    nonzero = mesh.area_faces > tol.zero
    # adjacencies with two nonzero faces
    adj_ok = nonzero[mesh.face_adjacency].all(axis=1)

    # if none of our face pairs are both nonzero exit
    # TODO : is this the correct check?
    # or should we just compare the projections
    # to the mesh scale
    if not adj_ok.any():
        return False

    # make threshold of convexity scale- relative
    threshold = tol.planar * mesh.scale

    # if projections of vertex onto plane of adjacent
    # face is negative, it means the face pair is locally
    # convex, and if that is true for all faces the mesh is convex
    convex = bool(mesh.face_adjacency_projections[adj_ok].max() < threshold)

    return convex


def hull_points(obj, qhull_options="QbB Pp"):
    """
    Try to extract a convex set of points from multiple input formats.

    Details on qhull options:
      http://www.qhull.org/html/qh-quick.htm#options

    Parameters
    ---------
    obj: Trimesh object
         (n,d) points
         (m,) Trimesh objects

    Returns
    --------
    points: (o,d) convex set of points
    """
    if hasattr(obj, "convex_hull"):
        return obj.convex_hull.vertices

    initial = np.asanyarray(obj, dtype=np.float64)
    if len(initial.shape) != 2:
        raise ValueError("points must be (n, dimension)!")
    hull = ConvexHull(initial, qhull_options=qhull_options)
    points = hull.points[hull.vertices]

    return points



"""
parent.py
-------------

The base class for Trimesh, PointCloud, and Scene objects
"""

import abc
import os
from dataclasses import dataclass

import numpy as np

from . import bounds, caching
from . import transformations as tf
from .caching import cache_decorator
from .constants import tol
from .resolvers import ResolverLike
from .typed import Any, ArrayLike, Dict, NDArray, Optional, float64
from .util import ABC


@dataclass
class LoadSource:
    """
    Save information about where a particular object was loaded from.
    """

    # a file-like object that can be accessed
    file_obj: Optional[Any] = None

    # a cleaned file type string, i.e. "stl"
    file_type: Optional[str] = None

    # if this was originally loaded from a file path
    # save it here so we can check it later.
    file_path: Optional[str] = None

    # did we open `file_obj` ourselves?
    was_opened: bool = False

    # a resolver for loading assets next to the file
    resolver: Optional[ResolverLike] = None

    @property
    def file_name(self) -> Optional[str]:
        """
        Get just the file name from the path if available.

        Returns
        ---------
        file_name
          Just the file name, i.e. for file_path="/a/b/c.stl" -> "c.stl"
        """
        if self.file_path is None:
            return None
        return os.path.basename(self.file_path)

    def __getstate__(self) -> Dict:
        # this overrides the `pickle.dump` behavior for this class
        # we cannot pickle a file object so return `file_obj: None` for pickles
        return {k: v if k != "file_obj" else None for k, v in self.__dict__.items()}

    def __deepcopy__(self, *args):
        return LoadSource(**self.__getstate__())


class Geometry(ABC):
    """
    `Geometry` is the parent class for all geometry.

    By decorating a method with `abc.abstractmethod` it means
    the objects that inherit from `Geometry` MUST implement
    those methods.
    """

    # geometry should have a dict to store loose metadata
    metadata: Dict

    @property
    def source(self) -> LoadSource:
        """
        Where and what was this current geometry loaded from?

        Returns
        --------
        source
          If loaded from a file, has the path, type, etc.
        """
        # this should have been tacked on by the loader
        # but we want to *always* be able to access
        # a value like `mesh.source.file_type` so add a default
        current = getattr(self, "_source", None)
        if current is not None:
            return current
        self._source = LoadSource()
        return self._source

    @property
    @abc.abstractmethod
    def identifier_hash(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def bounds(self) -> NDArray[np.float64]:
        pass

    @property
    @abc.abstractmethod
    def extents(self) -> NDArray[np.float64]:
        pass

    @abc.abstractmethod
    def apply_transform(self, matrix: ArrayLike) -> Any:
        pass

    @property
    @abc.abstractmethod
    def is_empty(self) -> bool:
        pass

    def __hash__(self):
        """
        Get a hash of the current geometry.

        Returns
        ---------
        hash
          Hash of current graph and geometry.
        """
        return self._data.__hash__()  # type: ignore

    @abc.abstractmethod
    def copy(self):
        pass

    @abc.abstractmethod
    def show(self):
        pass

    @abc.abstractmethod
    def __add__(self, other):
        pass

    @abc.abstractmethod
    def export(self, file_obj, file_type=None):
        pass

    def __repr__(self) -> str:
        """
        Print quick summary of the current geometry without
        computing properties.

        Returns
        -----------
        repr : str
          Human readable quick look at the geometry.
        """
        elements = []
        if hasattr(self, "vertices"):
            # for Trimesh and PointCloud
            elements.append(f"vertices.shape={self.vertices.shape}")
        if hasattr(self, "faces"):
            # for Trimesh
            elements.append(f"faces.shape={self.faces.shape}")
        if hasattr(self, "geometry") and isinstance(self.geometry, dict):
            # for Scene
            elements.append(f"len(geometry)={len(self.geometry)}")
        if "Voxel" in type(self).__name__:
            # for VoxelGrid objects
            elements.append(str(self.shape)[1:-1])
        if "file_name" in self.metadata:
            display = self.metadata["file_name"]
            elements.append(f"name=`{display}`")
        return "<trimesh.{}({})>".format(type(self).__name__, ", ".join(elements))

    def apply_translation(self, translation: ArrayLike):
        """
        Translate the current mesh.

        Parameters
        ----------
        translation : (3,) float
          Translation in XYZ
        """
        translation = np.asanyarray(translation, dtype=np.float64)
        if translation.shape == (2,):
            # create a planar matrix if we were passed a 2D offset
            return self.apply_transform(tf.planar_matrix(offset=translation))
        elif translation.shape != (3,):
            raise ValueError("Translation must be (3,) or (2,)!")

        # manually create a translation matrix
        matrix = np.eye(4)
        matrix[:3, 3] = translation
        return self.apply_transform(matrix)

    def apply_scale(self, scaling):
        """
        Scale the mesh.

        Parameters
        ----------
        scaling : float or (3,) float
          Scale factor to apply to the mesh
        """
        matrix = tf.scale_and_translate(scale=scaling)
        # apply_transform will work nicely even on negative scales
        return self.apply_transform(matrix)

    def __radd__(self, other):
        """
        Concatenate the geometry allowing concatenation with
        built in `sum()` function:
          `sum(Iterable[trimesh.Trimesh])`

        Parameters
        ------------
        other : Geometry
          Geometry or 0

        Returns
        ----------
        concat : Geometry
          Geometry of combined result
        """

        if other == 0:
            # adding 0 to a geometry never makes sense
            return self
        # otherwise just use the regular add function
        return self.__add__(type(self)(other))

    @cache_decorator
    def scale(self) -> float:
        """
        A loosely specified "order of magnitude scale" for the
        geometry which always returns a value and can be used
        to make code more robust to large scaling differences.

        It returns the diagonal of the axis aligned bounding box
        or if anything is invalid or undefined, `1.0`.

        Returns
        ----------
        scale : float
          Approximate order of magnitude scale of the geometry.
        """
        # if geometry is empty return 1.0
        if self.extents is None:
            return 1.0

        # get the length of the AABB diagonal
        scale = float((self.extents**2).sum() ** 0.5)
        if scale < tol.zero:
            return 1.0

        return scale

    @property
    def units(self) -> Optional[str]:
        """
        Definition of units for the mesh.

        Returns
        ----------
        units : str
          Unit system mesh is in, or None if not defined
        """
        return self.metadata.get("units", None)

    @units.setter
    def units(self, value: str) -> None:
        """
        Define the units of the current mesh.
        """
        self.metadata["units"] = str(value).lower().strip()


class Geometry3D(Geometry):
    """
    The `Geometry3D` object is the parent object of geometry objects
    which are three dimensional, including Trimesh, PointCloud,
    and Scene objects.
    """

    @caching.cache_decorator
    def bounding_box(self):
        """
        An axis aligned bounding box for the current mesh.

        Returns
        ----------
        aabb : trimesh.primitives.Box
          Box object with transform and extents defined
          representing the axis aligned bounding box of the mesh
        """
        from . import primitives

        transform = np.eye(4)
        # translate to center of axis aligned bounds
        transform[:3, 3] = self.bounds.mean(axis=0)

        return primitives.Box(transform=transform, extents=self.extents, mutable=False)

    @caching.cache_decorator
    def bounding_box_oriented(self):
        """
        An oriented bounding box for the current mesh.

        Returns
        ---------
        obb : trimesh.primitives.Box
          Box object with transform and extents defined
          representing the minimum volume oriented
          bounding box of the mesh
        """
        from . import bounds, primitives

        to_origin, extents = bounds.oriented_bounds(self)
        return primitives.Box(
            transform=np.linalg.inv(to_origin), extents=extents, mutable=False
        )

    @caching.cache_decorator
    def bounding_sphere(self):
        """
        A minimum volume bounding sphere for the current mesh.

        Note that the Sphere primitive returned has an unpadded
        exact `sphere_radius` so while the distance of every vertex
        of the current mesh from sphere_center will be less than
        sphere_radius, the faceted sphere primitive may not
        contain every vertex.

        Returns
        --------
        minball : trimesh.primitives.Sphere
          Sphere primitive containing current mesh
        """
        from . import nsphere, primitives

        center, radius = nsphere.minimum_nsphere(self)
        return primitives.Sphere(center=center, radius=radius, mutable=False)

    @caching.cache_decorator
    def bounding_cylinder(self):
        """
        A minimum volume bounding cylinder for the current mesh.

        Returns
        --------
        mincyl : trimesh.primitives.Cylinder
          Cylinder primitive containing current mesh
        """
        from . import bounds, primitives

        kwargs = bounds.minimum_cylinder(self)
        return primitives.Cylinder(mutable=False, **kwargs)

    @caching.cache_decorator
    def bounding_primitive(self):
        """
        The minimum volume primitive (box, sphere, or cylinder) that
        bounds the mesh.

        Returns
        ---------
        bounding_primitive : object
          Smallest primitive which bounds the mesh:
          trimesh.primitives.Sphere
          trimesh.primitives.Box
          trimesh.primitives.Cylinder
        """
        options = [
            self.bounding_box_oriented,
            self.bounding_sphere,
            self.bounding_cylinder,
        ]
        volume_min = np.argmin([i.volume for i in options])
        return options[volume_min]

    def apply_obb(self, **kwargs) -> NDArray[float64]:
        """
        Apply the oriented bounding box transform to the current mesh.

        This will result in a mesh with an AABB centered at the
        origin and the same dimensions as the OBB.

        Parameters
        ------------
        kwargs
          Passed through to `bounds.oriented_bounds`

        Returns
        ----------
        matrix : (4, 4) float
          Transformation matrix that was applied
          to mesh to move it into OBB frame
        """
        # save the pre-transform volume
        if tol.strict and hasattr(self, "volume"):
            volume = self.volume

        # calculate the OBB passing keyword arguments through
        matrix, extents = bounds.oriented_bounds(self, **kwargs)
        # apply the transform
        self.apply_transform(matrix)

        if tol.strict:
            # obb transform should not have changed volume
            if hasattr(self, "volume") and getattr(self, "is_watertight", False):
                assert np.isclose(self.volume, volume)
            # overall extents should match what we expected
            assert np.allclose(self.extents, extents)

        return matrix



import numpy as np

try:
    from . import generic as g
except BaseException:
    import generic as g


def test_scene():
    for file_name in ["cycloidal.ply", "sphere.ply"]:
        mesh = g.get_mesh(file_name)
        if mesh.units is None:
            mesh.units = "in"

        scene_split = g.trimesh.scene.split_scene(mesh)
        scene_split.convert_units("in")
        scene_base = g.trimesh.Scene(mesh)

        # save hash of scene before concat
        pre = [scene_split.__hash__(), scene_base.__hash__()]
        # make sure hash's give the same result twice
        assert scene_split.__hash__() == pre[0]
        assert scene_base.__hash__() == pre[1]

        # __hash__ is a long int which fails isinstance in Python 2
        assert type(scene_base.__hash__()).__name__ in ("int", "long")

        # try out scene appending
        concat = scene_split + scene_base

        # make sure concat didn't mess with original scenes
        assert scene_split.__hash__() == pre[0]
        assert scene_base.__hash__() == pre[1]

        # make sure concatenate appended things, stuff
        assert len(concat.geometry) == (
            len(scene_split.geometry) + len(scene_base.geometry)
        )

        for s in [scene_split, scene_base]:
            pre = s.__hash__()
            assert len(s.geometry) > 0
            assert s.is_valid

            flattened = s.graph.to_flattened()
            g.json.dumps(flattened)
            edgelist = s.graph.to_edgelist()
            g.json.dumps(edgelist)

            assert s.bounds.shape == (2, 3)
            assert s.centroid.shape == (3,)
            assert s.extents.shape == (3,)
            assert isinstance(s.scale, float)
            assert g.trimesh.util.is_shape(s.triangles, (-1, 3, 3))
            assert len(s.triangles) == len(s.triangles_node)

            assert s.__hash__() == pre
            assert s.__hash__() is not None

            # should be some duplicate nodes
            assert len(s.duplicate_nodes) > 0

            # should be a single scene camera
            assert isinstance(s.camera, g.trimesh.scene.cameras.Camera)
            # should be autogenerated lights
            assert len(s.lights) > 0
            # all lights should be lights
            assert all(isinstance(L, g.trimesh.scene.lighting.Light) for L in s.lights)
            # all lights should be added to scene graph
            assert all(L.name in s.graph for L in s.lights)

            # should have put a transform in scene graph
            assert s.camera.name in s.graph

            r = s.dump()

            gltf = s.export(file_type="gltf")
            assert isinstance(gltf, dict)
            assert len(gltf) > 0
            assert len(gltf["model.gltf"]) > 0

            glb = s.export(file_type="glb")
            assert len(glb) > 0
            assert isinstance(glb, bytes)

            for export_format in ["dict", "dict64"]:
                # try exporting the scene as a dict
                # then make sure json can serialize it
                e = g.json.dumps(s.export(file_type=export_format))
                # reconstitute the dict into a scene
                r = g.trimesh.load(g.json.loads(e))

                # make sure the extents are similar before and after
                assert g.np.allclose(g.np.prod(s.extents), g.np.prod(r.extents))

            # move the scene to origin
            s.rezero()
            # if our cache dump was bad this will fail
            assert g.np.allclose(s.centroid, 0, atol=1e-5)

            # make sure explode doesn't crash
            s.explode()


def test_cam_gltf():
    # Test that the camera is stored and loaded successfully into a Scene from a gltf.
    cam = g.trimesh.scene.cameras.Camera(fov=[60, 90], name="cam1")
    box = g.trimesh.creation.box(extents=[1, 2, 3])
    scene = g.trimesh.Scene(
        geometry=[box],
        camera=cam,
        camera_transform=np.array(
            [[0, 1, 0, -1], [1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
        ),
    )
    with g.TemporaryDirectory() as d:
        # exports by path allow files to be written
        path = g.os.path.join(d, "tmp.glb")
        scene.export(path)
        r = g.trimesh.load(path, force="scene")

        # ensure no added nodes
        assert set(r.graph.nodes) == {"world", "geometry_0", "cam1"}
        # ensure same camera parameters and extrinsics
        assert (r.camera_transform == scene.camera_transform).all()
        assert r.camera.name == cam.name
        assert (r.camera.fov == cam.fov).all()
        assert r.camera.z_near == cam.z_near


def test_scaling():
    # Test the scaling of scenes including unit conversion.

    scene = g.get_mesh("cycloidal.3DXML")

    hash_val = scene.__hash__()
    extents = scene.bounding_box_oriented.primitive.extents.copy()

    # TODO: have OBB return sorted extents
    # and adjust the transform to be correct

    factor = 10.0
    scaled = scene.scaled(factor)

    # the oriented bounding box should scale exactly
    # with the scaling factor
    assert g.np.allclose(scaled.bounding_box_oriented.primitive.extents / extents, factor)

    # check bounding primitives
    assert scene.bounding_box.volume > 0.0
    assert scene.bounding_primitive.volume > 0.0

    # we shouldn't have modified the original scene
    assert scene.__hash__() == hash_val
    assert scaled.__hash__() != hash_val

    # 3DXML comes in as mm
    assert all(m.units == "mm" for m in scene.geometry.values())
    assert scene.units == "mm"

    converted = scene.convert_units("in")

    assert g.np.allclose(
        converted.bounding_box_oriented.primitive.extents / extents,
        1.0 / 25.4,
        atol=1e-3,
    )

    # shouldn't have changed the original extents
    assert g.np.allclose(extents, scene.bounding_box_oriented.primitive.extents)

    # original shouldn't have changed
    assert converted.units == "in"
    assert all(m.units == "in" for m in converted.geometry.values())

    assert scene.units == "mm"

    # we shouldn't have modified the original scene
    assert scene.__hash__() == hash_val
    assert converted.__hash__() != hash_val


def test_scaling_3D():
    scene = g.get_mesh("cycloidal.3DXML")
    extents = scene.extents.copy()

    factor = [0.2, 1.3, 3.3]
    scaled = scene.scaled(factor)

    assert g.np.allclose(scaled.extents / extents, factor)

    factor = [3.0, 3.0, 3.0]
    scaled = scene.scaled(factor)

    assert g.np.allclose(scaled.extents / extents, factor)


def test_mixed_units():
    # create two boxes in a scene
    a = g.trimesh.creation.box()
    a.units = "in"

    b = g.trimesh.creation.box()
    b.units = "m"

    # mixed units should be None
    s = g.trimesh.Scene([a, b])
    assert len(s.geometry) == 2
    assert s.units is None

    # now all units should be meters and scene should report that
    a.units = "m"
    assert s.units == "m"


def test_scaling_3D_mixed():
    # same as test_scaling_3D but input scene contains 2D and 3D geometry
    scene = g.get_mesh("scenes.zip", mixed=True)
    extents = scene.extents.copy()

    factor = [0.2, 1.3, 3.3]
    scaled = scene.scaled(factor)

    assert g.np.allclose(scaled.extents / extents, factor)

    factor = [3.0, 3.0, 3.0]
    scaled = scene.scaled(factor)

    assert g.np.allclose(scaled.extents / extents, factor)


def test_add_geometry():
    # list-typed geometry should create multiple nodes,
    # e.g., below code is equivalent to
    #     scene.add_geometry(geometry[0], node_name='voxels')
    #     scene.add_geometry(geometry[1], parent_node_name='voxels')
    scene = g.trimesh.Scene()
    geometry = [g.trimesh.creation.box(), g.trimesh.creation.box()]
    scene.add_geometry(geometry)
    assert len(scene.graph.nodes_geometry) == 2


def test_add_concat():
    # create a scene with just a box in it
    a = g.trimesh.creation.box().scene()
    # do the same but move the box first
    b = g.trimesh.creation.box().apply_translation([2, 2, 2]).scene()
    # add the second scene to the first
    a.add_geometry(b)
    assert len(b.geometry) == 1
    assert len(a.geometry) == 2
    assert len(a.graph.nodes_geometry) == 2


def test_delete():
    # check to make sure our geometry delete cleans up
    a = g.trimesh.creation.icosphere()
    b = g.trimesh.creation.icosphere().apply_translation([2, 0, 0])
    s = g.trimesh.Scene({"a": a, "b": b})

    assert len(s.geometry) == 2
    assert len(s.graph.nodes_geometry) == 2
    # make sure every node has a transform
    [s.graph[n] for n in s.graph.nodes]

    # delete a geometry
    s.delete_geometry("a")
    assert len(s.geometry) == 1
    assert len(s.graph.nodes_geometry) == 1
    # if we screwed up the delete this will crash
    [s.graph[n] for n in s.graph.nodes]


def test_dupe():
    m = g.get_mesh("tube.obj", merge_norm=True, merge_tex=True)

    assert m.body_count == 1

    s = g.trimesh.scene.split_scene(m)
    assert len(s.graph.nodes) == 2
    assert len(s.graph.nodes_geometry) == 1
    assert len(s.duplicate_nodes) == 1
    assert len(s.duplicate_nodes[0]) == 1

    c = s.copy()
    assert len(c.graph.nodes) == 2
    assert len(c.graph.nodes_geometry) == 1
    assert len(c.duplicate_nodes) == 1
    assert len(c.duplicate_nodes[0]) == 1

    u = s.convert_units("in", guess=True)
    assert len(u.graph.nodes_geometry) == 1
    assert len(u.duplicate_nodes) == 1
    assert len(u.duplicate_nodes[0]) == 1


def test_3DXML():
    s = g.get_mesh("rod.3DXML")
    assert len(s.geometry) == 3
    assert len(s.graph.nodes_geometry) == 29

    dupe = s.duplicate_nodes
    assert len(dupe) == 3
    assert sum(len(i) for i in dupe) == 29

    # test cache dumping and survivability of bad
    # non-existent geometry specified in node_geometry
    s.graph.update(dupe[0][0], geometry="GARBAGE")
    # make sure geometry was updated
    assert s.graph[dupe[0][0]][1] == "GARBAGE"
    # get the regenerated duplicates
    dupe = s.duplicate_nodes
    assert len(dupe) == 3
    # should have been cleanly dropped
    assert sum(len(i) for i in dupe) == 28


def test_tri():
    scene = g.get_mesh("cycloidal.3DXML")

    # scene should have triangles
    assert g.trimesh.util.is_shape(scene.triangles, (-1, 3, 3))
    assert len(scene.triangles_node) == len(scene.triangles)

    # node name of inserted 2D geometry
    node = scene.add_geometry(g.get_mesh("2D/wrench.dxf"))
    # should be in the graph
    assert node in scene.graph.nodes
    # should have geometry defined
    assert node in scene.graph.nodes_geometry

    # 2D geometry has no triangles
    assert node not in scene.triangles_node
    # every geometry node except the one 2D thing
    # we inserted should be in triangles_node
    assert len(set(scene.triangles_node)) == len(scene.graph.nodes_geometry) - 1


def test_empty():
    m = g.get_mesh("bunny.ply")

    # not watertight so will result in empty scene
    s = g.trimesh.scene.split_scene(m)
    assert len(s.geometry) == 0

    s = s.convert_units("inches")
    n = s.duplicate_nodes
    assert len(n) == 0

    assert len(s.lights) == 0


def test_zipped():
    # Make sure a zip file with multiple file types
    # is returned as a single scene.

    # allow mixed 2D and 3D geometry
    m = g.get_mesh("scenes.zip", mixed=True)
    assert len(m.geometry) >= 6

    assert len(m.graph.nodes_geometry) >= 10
    assert any(isinstance(i, g.trimesh.path.Path2D) for i in m.geometry.values())
    assert any(isinstance(i, g.trimesh.Trimesh) for i in m.geometry.values())

    m = g.get_mesh("scenes.zip", mixed=False)
    assert len(m.geometry) < 6


def test_doubling():
    s = g.get_mesh("cycloidal.3DXML")

    # make sure we parked our car where we thought
    assert len(s.geometry) == 13

    # concatenate a scene with itself
    r = s + s

    # new scene should have twice as much geometry
    assert len(r.geometry) == (2 * len(s.geometry))

    assert g.np.allclose(s.extents, r.extents)

    # duplicate node groups should be twice as long
    set_ori = {len(i) * 2 for i in s.duplicate_nodes}
    set_dbl = {len(i) for i in r.duplicate_nodes}

    assert set_ori == set_dbl


def test_empty_scene():
    # test an empty scene
    empty = g.trimesh.Trimesh([], [])
    assert empty.bounds is None
    assert empty.extents is None
    assert g.np.isclose(empty.scale, 1.0)

    # create a sphere
    sphere = g.trimesh.creation.icosphere()

    # empty scene should have None for bounds
    scene = empty.scene()
    assert scene.bounds is None

    # add a sphere to the empty scene
    scene.add_geometry(sphere)
    # bounds should now be populated
    assert scene.bounds.shape == (2, 3)
    assert g.np.allclose(scene.bounds, sphere.bounds)


def test_transform():
    # check transforming scenes
    scene = g.trimesh.creation.box()
    assert g.np.allclose(scene.bounds, [[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]])

    scene.apply_translation([1, 0, 1])
    assert g.np.allclose(scene.bounds, [[0.5, -0.5, 0.5], [1.5, 0.5, 1.5]])


def test_material_group():
    # check scene is correctly grouped by materials
    s = g.get_mesh("box.obj", group_material=True)
    assert set(s.geometry.keys()) == {"Material", "SecondMaterial"}
    assert len(s.geometry["Material"].faces) == 8
    assert len(s.geometry["SecondMaterial"].faces) == 4

    # make sure our flag does something
    s = g.get_mesh("box.obj", group_material=False)
    assert set(s.geometry.keys()) != {"Material", "SecondMaterial"}


def test_strip():
    m = g.get_mesh("cycloidal.3DXML")
    assert any(g.visual.kind is not None for g in m.geometry.values())
    m.strip_visuals()
    assert all(g.visual.kind is None for g in m.geometry.values())


def test_export_concat():
    # Scenes exported in mesh formats should be
    # concatenating the meshes somewhere.
    original = g.trimesh.creation.icosphere(radius=0.123312)
    original_hash = original.identifier_hash

    scene = g.trimesh.Scene()
    scene.add_geometry(original)

    with g.TemporaryDirectory() as d:
        for ext in ["stl", "ply"]:
            file_name = g.os.path.join(d, "mesh." + ext)
            scene.export(file_name)
            loaded = g.trimesh.load(file_name)
            assert g.np.isclose(loaded.volume, original.volume)
    # nothing should have changed
    assert original.identifier_hash == original_hash


def test_exact_bounds():
    m = g.get_mesh("cycloidal.3DXML")
    assert isinstance(m, g.trimesh.Scene)

    dump = m.to_mesh()
    assert isinstance(dump, g.trimesh.Trimesh)

    # scene bounds should exactly match mesh bounds
    assert g.np.allclose(m.bounds, dump.bounds)


def test_concatenate_mixed():
    scene = g.trimesh.Scene(
        [
            g.trimesh.creation.icosphere(),
            g.trimesh.path.creation.rectangle([[0, 0], [1, 1]]),
        ]
    )

    dump = scene.to_mesh()
    assert isinstance(dump, g.trimesh.Trimesh)


def test_append_scenes():
    scene_0 = g.trimesh.Scene(base_frame="not_world")
    scene_1 = g.trimesh.Scene(base_frame="not_world")

    scene_sum = g.trimesh.scene.scene.append_scenes(
        (scene_0, scene_1), common=["not_world"], base_frame="not_world"
    )

    assert scene_sum.graph.base_frame == "not_world"


def test_scene_concat():
    # check that primitives get upgraded to meshes
    a = g.trimesh.Scene(
        [g.trimesh.primitives.Sphere(center=[5, 5, 5]), g.trimesh.primitives.Box()]
    )
    c = a.to_mesh()
    assert isinstance(c, g.trimesh.Trimesh)
    assert g.np.allclose(c.bounds, a.bounds)

    c = a.dump(concatenate=False)
    assert len(c) == len(a.geometry)

    # scene 2D
    scene_2D = g.trimesh.Scene(g.get_mesh("2D/250_cycloidal.DXF").split())
    concat = scene_2D.to_geometry()
    assert isinstance(concat, g.trimesh.path.Path2D)

    dump = scene_2D.dump(concatenate=False)
    assert len(dump) == len(scene_2D.geometry)
    assert all(isinstance(i, g.trimesh.path.Path2D) for i in dump)

    # all Path3D objects
    scene_3D = g.trimesh.Scene(
        [i.to_3D() for i in g.get_mesh("2D/250_cycloidal.DXF").split()]
    )

    dump = scene_3D.dump(concatenate=False)
    assert len(dump) >= 5
    assert all(isinstance(i, g.trimesh.path.Path3D) for i in dump)

    concat = scene_3D.to_geometry()
    assert isinstance(concat, g.trimesh.path.Path3D)

    mixed = list(scene_2D.geometry.values())
    mixed.extend(scene_3D.geometry.values())
    scene_mixed = g.trimesh.Scene(mixed)

    dump = scene_mixed.dump(concatenate=False)
    assert len(dump) == len(mixed)

    concat = scene_mixed.to_geometry()
    assert isinstance(concat, g.trimesh.path.Path3D)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    test_empty()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_edges():
    """
    Test edges_to_polygon
    """
    m = g.get_mesh("featuretype.STL")

    # get a polygon for the second largest facet
    index = m.facets_area.argsort()[-2]
    normal = m.facets_normal[index]
    origin = m._cache["facets_origin"][index]
    T = g.trimesh.geometry.plane_transform(origin, normal)
    vertices = g.trimesh.transform_points(m.vertices, T)[:, :2]

    # find boundary edges for the facet
    edges = m.edges_sorted.reshape((-1, 6))[m.facets[index]].reshape((-1, 2))
    group = g.trimesh.grouping.group_rows(edges, require_count=1)

    # run the polygon conversion
    polygon = g.trimesh.path.polygons.edges_to_polygons(
        edges=edges[group], vertices=vertices
    )

    assert len(polygon) == 1
    assert g.np.isclose(polygon[0].area, m.facets_area[index])

    # try transforming the polygon around
    M = g.np.eye(3)
    M[0][2] = 10.0
    P2 = g.trimesh.path.polygons.transform_polygon(polygon[0], M)
    distance = (
        g.np.array(P2.centroid.coords)[0] - g.np.array(polygon[0].centroid.coords)[0]
    )
    assert g.np.allclose(distance, [10.0, 0])


def test_random_polygon():
    """
    Test creation of random polygons
    """
    p = g.trimesh.path.polygons.random_polygon()
    assert p.area > 0.0
    assert p.is_valid


def test_polygon_sample():
    """
    Test random sampling of polygons
    """

    p = g.Point([0, 0]).buffer(1.0)
    count = 100

    s = g.trimesh.path.polygons.sample(p, count=count)
    assert len(s) <= count
    assert s.shape[1] == 2

    radius = (s**2).sum(axis=1).max()
    assert radius < (1.0 + 1e-8)

    # test Path2D sample wiring
    path = g.trimesh.load_path(p)
    s = path.sample(count=count)
    assert len(s) <= count
    assert s.shape[1] == 2
    radius = (s**2).sum(axis=1).max()
    assert radius < (1.0 + 1e-8)

    # try getting OBB of samples
    _T, extents = g.trimesh.path.polygons.polygon_obb(s)
    # OBB of samples should be less than diameter of circle
    diameter = g.np.ptp(g.np.reshape(p.bounds, (2, 2)), axis=0).max()
    assert (extents <= diameter).all()

    # test sampling with multiple bodies
    for i in range(3):
        assert g.np.isclose(path.area, p.area * (i + 1))
        path = path + g.trimesh.load_path(g.Point([(i + 2) * 2, 0]).buffer(1.0))
        s = path.sample(count=count)
        assert s.shape[1] == 2

    # try a polygon with a low area/aabb-area ratio to
    # check the iteration loop.
    p = g.Polygon([[0, 0], [1, 1], [0.5 - 1e-3, 0.5], [0, 0]])
    s = g.trimesh.path.polygons.sample(polygon=p, count=100)


def test_project():
    m = g.trimesh.creation.icosphere(subdivisions=4)

    p = [g.trimesh.path.polygons.projected(m, normal=n) for n in g.random((100, 3))]

    # sphere projection should never have interiors
    assert all(len(i.interiors) == 0 for i in p)
    # sphere projected area should always be close to pi
    assert g.np.allclose([i.area for i in p], g.np.pi, atol=0.05)


def test_project_backface():
    m = g.trimesh.Trimesh(vertices=[[0, 0, 0], [0, 1, 0], [1, 0, 0]], faces=[[0, 1, 2]])

    # check ignore_sign argument
    front = m.projected(m.face_normals[0], ignore_sign=False)
    assert len(front.entities) == 1
    front = m.projected(m.face_normals[0], ignore_sign=True)
    assert len(front.entities) == 1

    back = m.projected(-m.face_normals[0], ignore_sign=False)
    assert len(back.entities) == 0
    back = m.projected(-m.face_normals[0], ignore_sign=True)
    assert len(back.entities) == 1


def test_project_multi():
    for ignore_sign, precise in g.itertools.combinations([True, False] * 2, 2):
        mesh = g.trimesh.creation.box() + g.trimesh.creation.box().apply_translation(
            [3, 0, 0]
        )
        proj = mesh.projected(normal=[0, 0, 1], ignore_sign=ignore_sign, precise=precise)

        assert mesh.body_count == 2
        assert len(proj.root) == 2
        assert g.np.isclose(proj.area, 2.0)


def test_second_moment():
    def rectangle(extents):
        # return the boundary of an origin-centered
        # rectangle as a numpy array
        a = g.np.abs(g.np.array(extents) / 2.0)
        lower, upper = -a, a
        return g.np.array([lower, [upper[0], lower[1]], upper, [lower[0], upper[1]]])

    def poly(bh, bhi=None):
        # return a rectangle centered at the origin
        # as a shapely Polygon
        shell = rectangle(bh)
        if bhi is not None:
            holes = [rectangle(bhi)]
        else:
            holes = []
        return Polygon(shell=shell, holes=holes)

    def poly_corner(bh):
        # return a rectangle with one corner
        # at the origin and the rest in positive space
        shell = rectangle(bh)
        shell += shell.min(axis=0)
        return Polygon(shell=shell)

    def poly_doublecorner(bh):
        # returns two equal sized rectangles as one polygon
        # Same as poly_corner(), but the rectangle gets
        # mirrored by 180 deg at the origin
        # This puts the centroid in the origin with Ixy != 0
        shell_1 = rectangle(bh)
        shell_1 += shell_1.min(axis=0)
        shell_2 = -shell_1
        shell = g.np.concatenate(
            (shell_1[2:, :], shell_1[:2, :], shell_2[2:, :], shell_2[:2, :]), axis=0
        )
        return Polygon(shell=shell)

    def truth(bh, bhi=None):
        # return the analytical second moment of area
        # for a rectangle with centroid at the origin
        # and width-height of `bh` and an interior
        # rectangle width-height of `bhi`
        if bhi is None:
            bhi = g.np.zeros(2)
        b, h = bh
        bi, hi = bhi
        return (
            g.np.array(
                [b * h**3 - bi * hi**3, h * b**3 - hi * bi**3, 0.0],
                dtype=g.np.float64,
            )
            / 12
        )

    def truth_corner(bh):
        # check a rectangle with one corner
        # at the origin and the rest in positive space
        b, h = bh
        return g.np.array(
            [b * h**3 / 3.0, h * b**3 / 3.0, 0.5 * b**2 * 0.5 * h**2],
            dtype=g.np.float64,
        )

    from shapely.geometry import Polygon

    from trimesh.path.polygons import second_moments, transform_polygon

    heights = g.np.array([[0.01, 0.01], [1, 1], [10, 2], [3, 21]])
    for bh in heights:
        # check the second moment of a rectangle
        # as polygon is already centered, centered doesn't have any effect
        O_moments, O_principal_moments, O_alpha, O_transform = second_moments(
            poly(bh), return_centered=True
        )
        # check against wikipedia
        t = truth(bh)
        # for a centered rectangle, the principal axis are already aligned
        # with the frame axis
        assert g.np.allclose(O_moments, t)
        assert g.np.any(g.np.isclose(O_moments, O_principal_moments[0]))
        assert g.np.isclose(O_moments[2], 0)  # Ixy = 0
        assert g.np.isclose(O_alpha, 0)
        assert g.np.allclose(O_transform, g.np.eye(3))

        # now check a rectangle with the corner, so Ixy != 0

        # First we test with centering. The results should be same as
        # with the initially centered rectangles
        C_moments, C_principal_moments, C_alpha, C_transform = second_moments(
            poly_corner(bh), return_centered=True
        )
        assert g.np.allclose(O_moments, C_moments)
        assert g.np.allclose(O_principal_moments, C_principal_moments)
        assert g.np.isclose(O_alpha, C_alpha)
        assert g.np.allclose(O_transform[:, :2], C_transform[:, :2])

        # Now without centering
        moments = second_moments(poly_corner(bh), return_centered=False)
        t = truth_corner(bh)
        assert g.np.allclose(moments, t)

        # Now we will get the transform for a double rectangle. Then we will apply
        # the transform and test if Ixy == 0, alpha == 0 etc.
        C_moments, C_principal_moments, C_alpha, C_transform = second_moments(
            poly_doublecorner(bh), return_centered=True
        )
        # apply the outputted transform to the polygon
        T_polygon = transform_polygon(poly_doublecorner(bh), C_transform)
        # call the function on the transformed polygon
        T_moments, T_principal_moments, T_alpha, T_transform = second_moments(
            T_polygon, return_centered=True
        )
        assert g.np.any(g.np.isclose(T_moments, C_principal_moments[0]))
        assert g.np.allclose(C_principal_moments, T_principal_moments)
        assert g.np.isclose(T_alpha, 0, atol=1e-7)
        assert g.np.allclose(T_transform, g.np.eye(3))

        # check polygons with interior
        for bhi in heights:
            # only check if interior is smaller than exterior
            if not (bhi < bh).all():
                continue

            # check a rectangle with interiors
            c = second_moments(poly(bh, bhi), return_centered=False)
            t = truth(bh, bhi)
            assert g.np.allclose(c, t)


def test_native_centroid():
    # check our native implementation of the shoelace algorithm
    # against the results from shapely which we use for native
    # checks of counter-clockwise 2D coordinate loops

    # get some polygons without interiors
    polygons = [g.Polygon(i) for i in g.data["nestable"]]
    # same data as (n, 2) float arrays
    coords = [g.np.array(i) for i in g.data["nestable"]]

    for p, c in zip(polygons, coords):
        # area will be signed with respect to counter-clockwise
        _ccw, area, centroid = g.trimesh.util.is_ccw(c, return_all=True)
        assert g.np.allclose(centroid, g.np.array(p.centroid.coords)[0])
        assert g.np.isclose(abs(area), p.area)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    test_polygon_sample()



"""
boolean.py
-------------

Do boolean operations on meshes using either Blender or Manifold.
"""

import numpy as np

from . import interfaces
from .exceptions import ExceptionWrapper
from .iteration import reduce_cascade
from .typed import BooleanEngineType, BooleanOperationType, Callable, Dict, Sequence

try:
    from manifold3d import Manifold, Mesh
except BaseException as E:
    Mesh = ExceptionWrapper(E)
    Manifold = ExceptionWrapper(E)


def difference(
    meshes: Sequence,
    engine: BooleanEngineType = None,
    check_volume: bool = True,
    **kwargs,
):
    """
    Compute the boolean difference between a mesh an n other meshes.

    Parameters
    ----------
    meshes : sequence of trimesh.Trimesh
      Meshes to be processed.
    engine
      Which backend to use, i.e. 'blender' or 'manifold'
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.

    Returns
    ----------
    difference
      A `Trimesh` that contains `meshes[0] - meshes[1:]`
    """

    return _engines[engine](
        meshes, operation="difference", check_volume=check_volume, **kwargs
    )


def union(
    meshes: Sequence,
    engine: BooleanEngineType = None,
    check_volume: bool = True,
    **kwargs,
):
    """
    Compute the boolean union between a mesh an n other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.

    Returns
    ----------
    union
      A `Trimesh` that contains the union of all passed meshes.
    """
    return _engines[engine](
        meshes, operation="union", check_volume=check_volume, **kwargs
    )


def intersection(
    meshes: Sequence,
    engine: BooleanEngineType = None,
    check_volume: bool = True,
    **kwargs,
):
    """
    Compute the boolean intersection between a mesh and other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.

    Returns
    ----------
    intersection
      A `Trimesh` that contains the intersection geometry.
    """
    return _engines[engine](
        meshes, operation="intersection", check_volume=check_volume, **kwargs
    )


def boolean_manifold(
    meshes: Sequence,
    operation: BooleanOperationType,
    check_volume: bool = True,
    **kwargs,
):
    """
    Run an operation on a set of meshes using the Manifold engine.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    operation
      Which boolean operation to do.
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.
    """
    if check_volume and not all(m.is_volume for m in meshes):
        raise ValueError("Not all meshes are volumes!")

    # Convert to manifold meshes
    manifolds = [
        Manifold(
            mesh=Mesh(
                vert_properties=np.array(mesh.vertices, dtype=np.float32),
                tri_verts=np.array(mesh.faces, dtype=np.uint32),
            )
        )
        for mesh in meshes
    ]

    # Perform operations
    if operation == "difference":
        if len(meshes) < 2:
            raise ValueError("Difference only defined over two meshes.")
        elif len(meshes) == 2:
            # apply the single difference
            result_manifold = manifolds[0] - manifolds[1]
        elif len(meshes) > 2:
            # union all the meshes to be subtracted from the final result
            unioned = reduce_cascade(lambda a, b: a + b, manifolds[1:])
            # apply the difference
            result_manifold = manifolds[0] - unioned
    elif operation == "union":
        result_manifold = reduce_cascade(lambda a, b: a + b, manifolds)
    elif operation == "intersection":
        result_manifold = reduce_cascade(lambda a, b: a ^ b, manifolds)
    else:
        raise ValueError(f"Invalid boolean operation: '{operation}'")

    # Convert back to trimesh meshes
    from . import Trimesh

    result_mesh = result_manifold.to_mesh()

    return Trimesh(
        vertices=result_mesh.vert_properties, faces=result_mesh.tri_verts, process=False
    )


# which backend boolean engines do we have
_engines: Dict[str, Callable] = {}

if isinstance(Manifold, ExceptionWrapper):
    # manifold isn't available so use the import error
    _engines["manifold"] = Manifold
else:
    # manifold3d is the preferred option
    _engines["manifold"] = boolean_manifold


if interfaces.blender.exists:
    # we have `blender` in the path which  we can call with subprocess
    _engines["blender"] = interfaces.blender.boolean
else:
    # failing that add a helpful error message
    _engines["blender"] = ExceptionWrapper(ImportError("`blender` is not in `PATH`"))

# pick the first value that isn't an ExceptionWrapper.
_engines[None] = next(
    (v for v in _engines.values() if not isinstance(v, ExceptionWrapper)),
    ExceptionWrapper(
        ImportError("No boolean backend: `pip install manifold3d` or install `blender`")
    ),
)

engines_available = {
    k for k, v in _engines.items() if not isinstance(v, ExceptionWrapper)
}



try:
    from . import generic as g
except BaseException:
    import generic as g


class GraphTest(g.unittest.TestCase):
    def setUp(self):
        self.engines = []
        try:
            self.engines.append("scipy")
        except BaseException:
            pass
        try:
            self.engines.append("networkx")
        except BaseException:
            pass

    def test_soup(self):
        # a soup of random triangles, with no adjacent pairs
        soup = g.get_mesh("soup.stl")

        assert len(soup.face_adjacency) == 0
        assert len(soup.face_adjacency_radius) == 0
        assert len(soup.face_adjacency_edges) == 0
        assert len(soup.face_adjacency_convex) == 0
        assert len(soup.face_adjacency_unshared) == 0
        assert len(soup.face_adjacency_angles) == 0
        assert len(soup.facets) == 0

    def test_components(self):
        # a soup of random triangles, with no adjacent pairs
        soup = g.get_mesh("soup.stl")
        # a mesh with multiple watertight bodies
        mult = g.get_mesh("cycloidal.ply")
        # a mesh with a single watertight body
        sing = g.get_mesh("featuretype.STL")
        # mesh with a single tetrahedron
        tet = g.get_mesh("tet.ply")

        for engine in self.engines:
            # without requiring watertight the split should be into every face
            split = soup.split(only_watertight=False, engine=engine)
            assert len(split) == len(soup.faces)

            # with watertight there should be an empty list
            split = soup.split(only_watertight=True, engine=engine)
            assert len(split) == 0

            split = mult.split(only_watertight=False, engine=engine)
            assert len(split) >= 119

            split = mult.split(only_watertight=True, engine=engine)
            assert len(split) >= 117

            # random triangles should have no facets
            facets = g.trimesh.graph.facets(mesh=soup, engine=engine)
            assert len(facets) == 0

            facets = g.trimesh.graph.facets(mesh=mult, engine=engine)
            assert all(len(i) >= 2 for i in facets)
            assert len(facets) >= 8654

            split = sing.split(only_watertight=False, engine=engine)
            assert len(split) == 1
            assert split[0].is_watertight
            assert split[0].is_winding_consistent

            split = sing.split(only_watertight=True, engine=engine)
            assert len(split) == 1
            assert split[0].is_watertight
            assert split[0].is_winding_consistent

            # single tetrahedron
            assert tet.is_volume
            assert tet.body_count == 1
            # regardless of method or flag we should have one body result
            split = tet.split(only_watertight=True, engine=engine)
            assert len(split) == 1
            split = tet.split(only_watertight=False, engine=engine)
            assert len(split) == 1

    def test_vertex_adjacency_graph(self):
        f = g.trimesh.graph.vertex_adjacency_graph

        # a mesh with a single watertight body
        sing = g.get_mesh("featuretype.STL")
        vert_adj_g = f(sing)
        assert len(sing.vertices) == len(vert_adj_g)

    def test_engine_time(self):
        for mesh in g.get_meshes():
            tic = [g.time.time()]
            for engine in self.engines:
                mesh.split(engine=engine, only_watertight=False)
                g.trimesh.graph.facets(mesh=mesh, engine=engine)
                tic.append(g.time.time())

            diff = g.np.abs(g.np.diff(tic))
            if diff.min() > 0.0:
                diff /= diff.min()

            g.log.info(
                "graph engine on %s (scale %f sec):\n%s",
                mesh.source.file_name,
                diff.min(),
                str(g.np.column_stack((self.engines, diff))),
            )

    def test_smoothed(self):
        # Make sure smoothing is keeping the same number
        # of faces.

        for name in ["ADIS16480.STL", "featuretype.STL"]:
            mesh = g.get_mesh(name)
            assert len(mesh.faces) == len(mesh.smooth_shaded.faces)

    def test_engines(self):
        edges = g.np.arange(10).reshape((-1, 2))
        for i in range(0, 20):
            check_engines(nodes=g.np.arange(i), edges=edges)
        edges = g.np.column_stack((g.np.arange(1, 11), g.np.arange(0, 10)))
        for i in range(0, 20):
            check_engines(nodes=g.np.arange(i), edges=edges)

    def test_watertight(self):
        m = g.get_mesh("shared.STL")  # NOQA
        # assert m.is_watertight
        # assert m.is_winding_consistent
        # assert m.is_volume

    def test_traversals(self):
        # Test traversals (BFS+DFS)

        # generate some simple test data
        simple_nodes = g.np.arange(20)
        simple_edges = g.np.column_stack((simple_nodes[:-1], simple_nodes[1:]))
        simple_edges = g.np.vstack(
            (simple_edges, [[19, 0], [10, 1000], [500, 501]])
        ).astype(g.np.int64)

        all_edges = g.data["edges"]
        all_edges.append(simple_edges)

        for edges in all_edges:
            edges = g.np.array(edges, dtype=g.np.int64)
            assert g.trimesh.util.is_shape(edges, (-1, 2))

            # collect the new nodes
            nodes = g.np.unique(edges)

            # the basic BFS/DFS traversal
            dfs_basic = g.trimesh.graph.traversals(edges, "dfs")
            bfs_basic = g.trimesh.graph.traversals(edges, "bfs")
            # check return types
            assert all(i.dtype == g.np.int64 for i in dfs_basic)
            assert all(i.dtype == g.np.int64 for i in bfs_basic)

            # check to make sure traversals visited every node
            dfs_set = set(g.np.hstack(dfs_basic))
            bfs_set = set(g.np.hstack(bfs_basic))
            nodes_set = set(nodes)
            assert dfs_set == nodes_set
            assert bfs_set == nodes_set

            # check traversal filling
            # fill_traversals should always include every edge
            # regardless of the path so test on bfs/dfs/empty
            for traversal in [dfs_basic, bfs_basic, []]:
                # disconnect consecutive nodes that are not edges
                # and add edges that were left off by jumps
                dfs = g.trimesh.graph.fill_traversals(traversal, edges)
                # edges that are included in the new separated traversal
                inc = g.trimesh.util.vstack_empty(
                    [g.np.column_stack((i[:-1], i[1:])) for i in dfs]
                )

                # make a set from edges included in the traversal
                inc_set = {
                    i.tobytes()
                    for i in g.trimesh.grouping.hashable_rows(g.np.sort(inc, axis=1))
                }
                # make a set of the source edges we were supposed to include
                edge_set = {
                    i.tobytes()
                    for i in g.trimesh.grouping.hashable_rows(g.np.sort(edges, axis=1))
                }

                # we should have exactly the same edges
                # after the filled traversal as we started with
                assert len(inc) == len(edges)
                # every edge should occur exactly once
                assert len(inc_set) == len(inc)
                # unique edges should be the same
                assert inc_set == edge_set

                # check all return dtypes
                assert all(i.dtype == g.np.int64 for i in dfs)

    def test_adjacency(self):
        for add_degen in [False, True]:
            for name in ["featuretype.STL", "soup.stl"]:
                m = g.get_mesh(name)
                if add_degen:
                    # make the first face degenerate
                    m.faces[0][2] = m.faces[0][0]
                # degenerate faces should be filtered
                assert g.np.not_equal(*m.face_adjacency.T).all()

                # check the various paths of calling face adjacency
                a = g.trimesh.graph.face_adjacency(
                    m.faces.view(g.np.ndarray).copy(), return_edges=False
                )
                b, be = g.trimesh.graph.face_adjacency(
                    m.faces.view(g.np.ndarray).copy(), return_edges=True
                )
                c = g.trimesh.graph.face_adjacency(mesh=m, return_edges=False)
                c, ce = g.trimesh.graph.face_adjacency(mesh=m, return_edges=True)
                # make sure they all return the expected result
                assert g.np.allclose(a, b)
                assert g.np.allclose(a, c)
                assert len(be) == len(a)
                assert len(ce) == len(a)

                # package properties to loop through
                zips = zip(
                    m.face_adjacency, m.face_adjacency_edges, m.face_adjacency_unshared
                )
                for a, e, v in zips:
                    # get two adjacenct faces as a set
                    fa = set(m.faces[a[0]])
                    fb = set(m.faces[a[1]])

                    # face should be different
                    assert fa != fb
                    # shared edge should be in both faces

                    # removing 2 vertices should leave one
                    da = fa.difference(e)
                    db = fb.difference(e)
                    assert len(da) == 1
                    assert len(db) == 1

                    # unshared vertex should be correct
                    assert da.issubset(v)
                    assert db.issubset(v)
                    assert da != db
                    assert len(v) == 2


def check_engines(edges, nodes):
    """
    Make sure connected component graph engines are
    returning the exact same values
    """
    results = []
    engines = [None, "scipy", "networkx"]

    for engine in engines:
        c = g.trimesh.graph.connected_components(edges, nodes=nodes, engine=engine)
        if len(c) > 0:
            # check to see if every resulting component
            # was in the passed set of nodes
            diff = g.np.setdiff1d(g.np.hstack(c), nodes)
            assert len(diff) == 0
        # store the result as a set of tuples so we can compare
        results.append({tuple(sorted(i)) for i in c})

    # make sure different engines are returning the same thing
    try:
        assert all(i == results[0] for i in results[1:])
    except BaseException as E:
        g.log.debug(results)
        raise E


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



from collections import defaultdict

import numpy as np

from ... import grouping, resources, util
from ... import transformations as tf
from ...constants import log
from ...constants import tol_path as tol
from ...util import multi_dict
from ..arc import to_threepoint
from ..entities import Arc, BSpline, Line, Text

# unit codes
_DXF_UNITS = {
    1: "inches",
    2: "feet",
    3: "miles",
    4: "millimeters",
    5: "centimeters",
    6: "meters",
    7: "kilometers",
    8: "microinches",
    9: "mils",
    10: "yards",
    11: "angstroms",
    12: "nanometers",
    13: "microns",
    14: "decimeters",
    15: "decameters",
    16: "hectometers",
    17: "gigameters",
    18: "AU",
    19: "light years",
    20: "parsecs",
}
# backwards, for reference
_UNITS_TO_DXF = {v: k for k, v in _DXF_UNITS.items()}

# a string which we will replace spaces with temporarily
_SAFESPACE = "|<^>|"

# save metadata to a DXF Xrecord starting here
# Valid values are 1-369 (except 5 and 105)
XRECORD_METADATA = 134
# the sentinel string for trimesh metadata
# this should be seen at XRECORD_METADATA
XRECORD_SENTINEL = "TRIMESH_METADATA:"
# the maximum line length before we split lines
XRECORD_MAX_LINE = 200
# the maximum index of XRECORDS
XRECORD_MAX_INDEX = 368


def load_dxf(file_obj, **kwargs):
    """
    Load a DXF file to a dictionary containing vertices and
    entities.

    Parameters
    ----------
    file_obj: file or file- like object (has object.read method)

    Returns
    ----------
    result: dict, keys are  entities, vertices and metadata
    """

    # in a DXF file, lines come in pairs,
    # a group code then the next line is the value
    # we are removing all whitespace then splitting with the
    # splitlines function which uses the universal newline method
    raw = file_obj.read()
    # if we've been passed bytes
    if hasattr(raw, "decode"):
        # search for the sentinel string indicating binary DXF
        # do it by encoding sentinel to bytes and subset searching
        if raw[:22].find(b"AutoCAD Binary DXF") != -1:
            # no converter to ASCII DXF available
            raise NotImplementedError("Binary DXF is not supported!")
        else:
            # we've been passed bytes that don't have the
            # header for binary DXF so try decoding as UTF-8
            raw = raw.decode("utf-8", errors="ignore")

    # remove trailing whitespace
    raw = str(raw).strip()
    # without any spaces and in upper case
    cleaned = raw.replace(" ", "").strip().upper()

    # blob with spaces and original case
    blob_raw = np.array(str.splitlines(raw)).reshape((-1, 2))
    # if this reshape fails, it means the DXF is malformed
    blob = np.array(str.splitlines(cleaned)).reshape((-1, 2))

    # get the section which contains the header in the DXF file
    endsec = np.nonzero(blob[:, 1] == "ENDSEC")[0]

    # store metadata
    metadata = {}

    # try reading the header, which may be malformed
    header_start = np.nonzero(blob[:, 1] == "HEADER")[0]
    if len(header_start) > 0:
        header_end = endsec[np.searchsorted(endsec, header_start[0])]
        header_blob = blob[header_start[0] : header_end]

        # store some properties from the DXF header
        metadata["DXF_HEADER"] = {}
        for key, group in [
            ("$ACADVER", "1"),
            ("$DIMSCALE", "40"),
            ("$DIMALT", "70"),
            ("$DIMALTF", "40"),
            ("$DIMUNIT", "70"),
            ("$INSUNITS", "70"),
            ("$LUNITS", "70"),
        ]:
            value = get_key(header_blob, key, group)
            if value is not None:
                metadata["DXF_HEADER"][key] = value

        # store unit data pulled from the header of the DXF
        # prefer LUNITS over INSUNITS
        # I couldn't find a table for LUNITS values but they
        # look like they are 0- indexed versions of
        # the INSUNITS keys, so for now offset the key value
        for offset, key in [(-1, "$LUNITS"), (0, "$INSUNITS")]:
            # get the key from the header blob
            units = get_key(header_blob, key, "70")
            # if it exists add the offset
            if units is None:
                continue
            metadata[key] = units
            units += offset
            # if the key is in our list of units store it
            if units in _DXF_UNITS:
                metadata["units"] = _DXF_UNITS[units]
        # warn on drawings with no units
        if "units" not in metadata:
            log.debug("DXF doesn't have units specified!")

    # get the section which contains entities in the DXF file
    entity_start = np.nonzero(blob[:, 1] == "ENTITIES")[0][0]
    entity_end = endsec[np.searchsorted(endsec, entity_start)]

    blocks = None
    check_entity = blob[entity_start:entity_end][:, 1]
    # only load blocks if an entity references them via an INSERT
    if "INSERT" in check_entity or "BLOCK" in check_entity:
        try:
            # which part of the raw file contains blocks
            block_start = np.nonzero(blob[:, 1] == "BLOCKS")[0][0]
            block_end = endsec[np.searchsorted(endsec, block_start)]

            blob_block = blob[block_start:block_end]
            blob_block_raw = blob_raw[block_start:block_end]
            block_infl = np.nonzero((blob_block == ["0", "BLOCK"]).all(axis=1))[0]

            # collect blocks by name
            blocks = {}
            for index in np.array_split(np.arange(len(blob_block)), block_infl):
                try:
                    v, e, name = convert_entities(
                        blob_block[index], blob_block_raw[index], return_name=True
                    )
                    if len(e) > 0:
                        blocks[name] = (v, e)
                except BaseException:
                    pass
        except BaseException:
            log.error("failed to parse blocks!", exc_info=True)

    # actually load referenced entities
    vertices, entities = convert_entities(
        blob[entity_start:entity_end], blob_raw[entity_start:entity_end], blocks=blocks
    )

    # return result as kwargs for trimesh.path.Path2D constructor
    result = {"vertices": vertices, "entities": entities, "metadata": metadata}

    return result


def convert_entities(blob, blob_raw=None, blocks=None, return_name=False):
    """
    Convert a chunk of entities into trimesh entities.

    Parameters
    ------------
    blob : (n, 2) str
      Blob of entities uppercased
    blob_raw : (n, 2) str
      Blob of entities not uppercased
    blocks : None or dict
      Blocks referenced by INSERT entities
    return_name : bool
      If True return the first '2' value

    Returns
    ----------
    """

    if blob_raw is None:
        blob_raw = blob

    def info(e):
        """
        Pull metadata based on group code, and return as a dict.
        """
        # which keys should we extract from the entity data
        # DXF group code : our metadata key
        get = {"8": "layer", "2": "name"}
        # replace group codes with names and only
        # take info from the entity dict if it is in cand
        renamed = {get[k]: util.make_sequence(v)[0] for k, v in e.items() if k in get}
        return renamed

    def convert_line(e):
        """
        Convert DXF LINE entities into trimesh Line entities.
        """
        # create a single Line entity
        entities.append(Line(points=len(vertices) + np.arange(2), **info(e)))
        # add the vertices to our collection
        vertices.extend(
            np.array([[e["10"], e["20"]], [e["11"], e["21"]]], dtype=np.float64)
        )

    def convert_circle(e):
        """
        Convert DXF CIRCLE entities into trimesh Circle entities
        """
        R = float(e["40"])
        C = np.array([e["10"], e["20"]]).astype(np.float64)
        points = to_threepoint(center=C[:2], radius=R)
        entities.append(
            Arc(points=(len(vertices) + np.arange(3)), closed=True, **info(e))
        )
        vertices.extend(points)

    def convert_arc(e):
        """
        Convert DXF ARC entities into into trimesh Arc entities.
        """
        # the radius of the circle
        R = float(e["40"])
        # the center point of the circle
        C = np.array([e["10"], e["20"]], dtype=np.float64)
        # the start and end angle of the arc, in degrees
        # this may depend on an AUNITS header data
        A = np.radians(np.array([e["50"], e["51"]], dtype=np.float64))
        # convert center/radius/angle representation
        # to three points on the arc representation
        points = to_threepoint(center=C[:2], radius=R, angles=A)
        # add a single Arc entity
        entities.append(Arc(points=len(vertices) + np.arange(3), closed=False, **info(e)))
        # add the three vertices
        vertices.extend(points)

    def convert_polyline(e):
        """
        Convert DXF LWPOLYLINE entities into trimesh Line entities.
        """
        # load the points in the line
        lines = np.column_stack((e["10"], e["20"])).astype(np.float64)

        # save entity info so we don't have to recompute
        polyinfo = info(e)

        # 70 is the closed flag for polylines
        # if the closed flag is set make sure to close
        is_closed = "70" in e and int(e["70"][0]) & 1
        if is_closed:
            lines = np.vstack((lines, lines[:1]))

        # 42 is the vertex bulge flag for LWPOLYLINE entities
        # "bulge" is autocad for "add a stupid arc using flags
        # in my otherwise normal polygon", it's like SVG arc
        # flags but somehow even more annoying
        if "42" in e:
            # get the actual bulge float values
            bulge = np.array(e["42"], dtype=np.float64)
            # what position were vertices stored at
            vid = np.nonzero(chunk[:, 0] == "10")[0]
            # what position were bulges stored at in the chunk
            bid = np.nonzero(chunk[:, 0] == "42")[0]
            # filter out endpoint bulge if we're not closed
            if not is_closed:
                bid_ok = bid < vid.max()
                bid = bid[bid_ok]
                bulge = bulge[bid_ok]
            # which vertex index is bulge value associated with
            bulge_idx = np.searchsorted(vid, bid)
            # convert stupid bulge to Line/Arc entities
            v, e = bulge_to_arcs(
                lines=lines, bulge=bulge, bulge_idx=bulge_idx, is_closed=is_closed
            )
            for i in e:
                # offset added entities by current vertices length
                i.points += len(vertices)
            vertices.extend(v)
            entities.extend(e)
            # done with this polyline
            return

        # we have a normal polyline so just add it
        # as single line entity and vertices
        entities.append(Line(points=np.arange(len(lines)) + len(vertices), **polyinfo))
        vertices.extend(lines)

    def convert_bspline(e):
        """
        Convert DXF Spline entities into trimesh BSpline entities.
        """
        # in the DXF there are n points and n ordered fields
        # with the same group code

        points = np.column_stack((e["10"], e["20"])).astype(np.float64)
        knots = np.array(e["40"]).astype(np.float64)

        # if there are only two points, save it as a line
        if len(points) == 2:
            # create a single Line entity
            entities.append(Line(points=len(vertices) + np.arange(2), **info(e)))
            # add the vertices to our collection
            vertices.extend(points)
            return

        # check bit coded flag for closed
        # closed = bool(int(e['70'][0]) & 1)
        # check euclidean distance to see if closed
        closed = np.linalg.norm(points[0] - points[-1]) < tol.merge

        # create a BSpline entity
        entities.append(
            BSpline(
                points=np.arange(len(points)) + len(vertices),
                knots=knots,
                closed=closed,
                **info(e),
            )
        )
        # add the vertices
        vertices.extend(points)

    def convert_text(e):
        """
        Convert a DXF TEXT entity into a native text entity.
        """
        # text with leading and trailing whitespace removed
        text = e["1"].strip()
        # try getting optional height of text
        try:
            height = float(e["40"])
        except BaseException:
            height = None
        try:
            # rotation angle converted to radians
            angle = np.radians(float(e["50"]))
        except BaseException:
            # otherwise no rotation
            angle = 0.0
        # origin point
        origin = np.array([e["10"], e["20"]], dtype=np.float64)
        # an origin-relative point (so transforms work)
        vector = origin + [np.cos(angle), np.sin(angle)]
        # try to extract a (horizontal, vertical) text alignment
        align = ["center", "center"]
        try:
            align[0] = ["left", "center", "right"][int(e["72"])]
        except BaseException:
            pass
        # append the entity
        entities.append(
            Text(
                origin=len(vertices),
                vector=len(vertices) + 1,
                height=height,
                text=text,
                align=align,
            )
        )
        # append the text origin and direction
        vertices.append(origin)
        vertices.append(vector)

    def convert_insert(e):
        """
        Convert an INSERT entity, which inserts a named group of
        entities (i.e. a "BLOCK") at a specific location.
        """
        if blocks is None:
            return

        # name of block to insert
        name = e["2"]
        # if we haven't loaded the block skip
        if name not in blocks:
            return
        # angle to rotate the block by
        angle = float(e.get("50", 0.0))
        # the insertion point of the block
        offset = np.array([e.get("10", 0.0), e.get("20", 0.0)], dtype=np.float64)
        # what to scale the block by
        scale = np.array([e.get("41", 1.0), e.get("42", 1.0)], dtype=np.float64)

        # the current entities and vertices of the referenced block.
        cv, ce = blocks[name]
        for i in ce:
            # copy the referenced entity as it may be included multiple times
            entities.append(i.copy())
            # offset its vertices to the current index
            entities[-1].points += len(vertices)
        # transform the block's vertices based on the entity settings
        vertices.extend(
            tf.transform_points(
                cv, tf.planar_matrix(offset=offset, theta=np.radians(angle), scale=scale)
            )
        )

    # find the start points of entities
    # DXF object to trimesh object converters
    loaders = {
        "LINE": (dict, convert_line),
        "LWPOLYLINE": (multi_dict, convert_polyline),
        "ARC": (dict, convert_arc),
        "CIRCLE": (dict, convert_circle),
        "SPLINE": (multi_dict, convert_bspline),
        "INSERT": (dict, convert_insert),
        "BLOCK": (dict, convert_insert),
    }

    # store loaded vertices
    vertices = []
    # store loaded entities
    entities = []
    # an old-style polyline entity strings its data across
    # multiple vertex entities like a real asshole
    polyline = None
    # chunks of entities are divided by group-code-0
    inflection = np.nonzero(blob[:, 0] == "0")[0]

    unsupported = defaultdict(lambda: 0)

    # loop through chunks of entity information
    for index in np.array_split(np.arange(len(blob)), inflection):
        # if there is only a header continue
        if len(index) < 1:
            continue
        # chunk will be an (n, 2) array of (group code, data) pairs
        chunk = blob[index]
        # the string representing entity type
        entity_type = chunk[0][1]

        # if we are referencing a block or insert by name make
        # sure the name key is in the original case vs upper-case
        if entity_type in ("BLOCK", "INSERT"):
            try:
                index_name = next(i for i, v in enumerate(chunk) if v[0] == "2")
                chunk[index_name][1] = blob_raw[index][index_name][1]
            except StopIteration:
                pass

        # special case old- style polyline entities
        if entity_type == "POLYLINE":
            polyline = [dict(chunk)]
        # if we are collecting vertex entities
        elif polyline is not None and entity_type == "VERTEX":
            polyline.append(dict(chunk))
        # the end of a polyline
        elif polyline is not None and entity_type == "SEQEND":
            # pull the geometry information for the entity
            lines = np.array([[i["10"], i["20"]] for i in polyline[1:]], dtype=np.float64)

            is_closed = False
            # check for a closed flag on the polyline
            if "70" in polyline[0]:
                # flag is bit- coded integer
                flag = int(polyline[0]["70"])
                # first bit represents closed
                is_closed = bool(flag & 1)
                if is_closed:
                    lines = np.vstack((lines, lines[:1]))

            # get the index of each bulged vertices
            bulge_idx = np.array(
                [i for i, e in enumerate(polyline) if "42" in e], dtype=np.int64
            )
            # get the actual bulge value
            bulge = np.array(
                [float(e["42"]) for i, e in enumerate(polyline) if "42" in e],
                dtype=np.float64,
            )
            # convert bulge to new entities
            cv, ce = bulge_to_arcs(
                lines=lines, bulge=bulge, bulge_idx=bulge_idx, is_closed=is_closed
            )
            for i in ce:
                # offset entities by existing vertices
                i.points += len(vertices)
            vertices.extend(cv)
            entities.extend(ce)
            # we no longer have an active polyline
            polyline = None
        elif entity_type == "TEXT":
            # text entities need spaces preserved so take
            # group codes from clean representation (0- column)
            # and data from the raw representation (1- column)
            chunk_raw = blob_raw[index]
            # if we didn't use clean group codes we wouldn't
            # be able to access them by key as whitespace
            # is random and crazy, like: '  1 '
            chunk_raw[:, 0] = blob[index][:, 0]
            try:
                convert_text(dict(chunk_raw))
            except BaseException:
                log.debug("failed to load text entity!", exc_info=True)
        # if the entity contains all relevant data we can
        # cleanly load it from inside a single function
        elif entity_type in loaders:
            # the chunker converts an (n,2) list into a dict
            chunker, loader = loaders[entity_type]
            # convert data to dict
            entity_data = chunker(chunk)
            # append data to the lists we're collecting
            loader(entity_data)
        elif entity_type != "ENTITIES":
            unsupported[entity_type] += 1
    if len(unsupported) > 0:
        log.debug(
            "skipping dxf entities: {}".format(
                ", ".join(f"{k}: {v}" for k, v in unsupported.items())
            )
        )
    # stack vertices into single array
    vertices = util.vstack_empty(vertices).astype(np.float64)
    if return_name:
        name = blob_raw[blob[:, 0] == "2"][0][1]
        return vertices, entities, name

    return vertices, entities


def export_dxf(path, only_layers=None):
    """
    Export a 2D path object to a DXF file.

    Parameters
    ----------
    path : trimesh.path.path.Path2D
      Input geometry to export
    only_layers : None or set
      If passed only export the layers specified

    Returns
    ----------
    export : str
      Path formatted as a DXF file
    """
    # get the template for exporting DXF files
    template = resources.get_json("templates/dxf.json")

    def format_points(points, as_2D=False, increment=True):
        """
        Format points into DXF- style point string.

        Parameters
        -----------
        points : (n,2) or (n,3) float
          Points in space
        as_2D : bool
          If True only output 2 points per vertex
        increment : bool
          If True increment group code per point
          Example:
            [[X0, Y0, Z0], [X1, Y1, Z1]]
          Result, new lines replaced with spaces:
            True  -> 10 X0 20 Y0 30 Z0 11 X1 21 Y1 31 Z1
            False -> 10 X0 20 Y0 30 Z0 10 X1 20 Y1 30 Z1

        Returns
        -----------
        packed : str
          Points formatted with group code
        """
        points = np.asanyarray(points, dtype=np.float64)
        # get points in 3D
        three = util.stack_3D(points)
        if increment:
            group = np.tile(
                np.arange(len(three), dtype=np.int64).reshape((-1, 1)), (1, 3)
            )
        else:
            group = np.zeros((len(three), 3), dtype=np.int64)
        group += [10, 20, 30]

        if as_2D:
            group = group[:, :2]
            three = three[:, :2]
        # join into result string
        packed = "\n".join(
            f"{g:d}\n{v:.12g}" for g, v in zip(group.reshape(-1), three.reshape(-1))
        )

        return packed

    def entity_info(entity):
        """
        Pull layer, color, and name information about an entity

        Parameters
        -----------
        entity : entity object
          Source entity to pull metadata

        Returns
        ----------
        subs : dict
          Has keys 'COLOR', 'LAYER', 'NAME'
        """
        # TODO : convert RGBA entity.color to index
        subs = {
            "COLOR": 255,  # default is ByLayer
            "LAYER": 0,
            "NAME": str(id(entity))[:16],
        }
        if hasattr(entity, "layer"):
            # make sure layer name is forced into ASCII
            subs["LAYER"] = util.to_ascii(entity.layer)
        return subs

    def convert_line(line, vertices):
        """
        Convert an entity to a discrete polyline

        Parameters
        -------------
        line : entity
          Entity which has 'e.discrete' method
        vertices : (n, 2) float
          Vertices in space

        Returns
        -----------
        as_dxf : str
          Entity exported as a DXF
        """
        # get a discrete representation of entity
        points = line.discrete(vertices)
        # if one or fewer points return nothing
        if len(points) <= 1:
            return ""

        # generate a substitution dictionary for template
        subs = entity_info(line)
        subs["POINTS"] = format_points(points, as_2D=True, increment=False)
        subs["TYPE"] = "LWPOLYLINE"
        subs["VCOUNT"] = len(points)
        # 1 is closed
        # 0 is default (open)
        subs["FLAG"] = int(bool(line.closed))

        result = template["line"].format(**subs)
        return result

    def convert_arc(arc, vertices):
        # get the center of arc and include span angles
        info = arc.center(vertices, return_angle=True, return_normal=False)
        subs = entity_info(arc)
        center = info.center
        if len(center) == 2:
            center = np.append(center, 0.0)
        data = "10\n{:.12g}\n20\n{:.12g}\n30\n{:.12g}".format(*center)
        data += f"\n40\n{info.radius:.12g}"

        if arc.closed:
            subs["TYPE"] = "CIRCLE"
        else:
            subs["TYPE"] = "ARC"
            # an arc is the same as a circle, with an added start
            # and end angle field
            data += "\n100\nAcDbArc"
            data += "\n50\n{:.12g}\n51\n{:.12g}".format(*np.degrees(info.angles))
        subs["DATA"] = data
        result = template["arc"].format(**subs)

        return result

    def convert_bspline(spline, vertices):
        # points formatted with group code
        points = format_points(vertices[spline.points], increment=False)

        # (n,) float knots, formatted with group code
        knots = ("40\n{:.12g}\n" * len(spline.knots)).format(*spline.knots)[:-1]

        # bit coded
        flags = {"closed": 1, "periodic": 2, "rational": 4, "planar": 8, "linear": 16}

        flag = flags["planar"]
        if spline.closed:
            flag = flag | flags["closed"]

        normal = [0.0, 0.0, 1.0]
        n_code = [210, 220, 230]
        n_str = "\n".join(f"{i:d}\n{j:.12g}" for i, j in zip(n_code, normal))

        subs = entity_info(spline)
        subs.update(
            {
                "TYPE": "SPLINE",
                "POINTS": points,
                "KNOTS": knots,
                "NORMAL": n_str,
                "DEGREE": 3,
                "FLAG": flag,
                "FCOUNT": 0,
                "KCOUNT": len(spline.knots),
                "PCOUNT": len(spline.points),
            }
        )
        # format into string template
        result = template["bspline"].format(**subs)

        return result

    def convert_text(txt, vertices):
        """
        Convert a Text entity to DXF string.
        """
        # start with layer info
        sub = entity_info(txt)
        # get the origin point of the text
        sub["ORIGIN"] = format_points(vertices[[txt.origin]], increment=False)
        # rotation angle in degrees
        sub["ANGLE"] = np.degrees(txt.angle(vertices))
        # actual string of text with spaces escaped
        # force into ASCII to avoid weird encoding issues
        sub["TEXT"] = (
            txt.text.replace(" ", _SAFESPACE)
            .encode("ascii", errors="ignore")
            .decode("ascii")
        )
        # height of text
        sub["HEIGHT"] = txt.height
        result = template["text"].format(**sub)
        return result

    def convert_generic(entity, vertices):
        """
        For entities we don't know how to handle, return their
        discrete form as a polyline
        """
        return convert_line(entity, vertices)

    # make sure we're not losing a ton of
    # precision in the string conversion
    np.set_printoptions(precision=12)
    # trimesh entity to DXF entity converters
    conversions = {
        "Line": convert_line,
        "Text": convert_text,
        "Arc": convert_arc,
        "Bezier": convert_generic,
        "BSpline": convert_bspline,
    }
    collected = []
    for e, layer in zip(path.entities, path.layers):
        name = type(e).__name__
        # only export specified layers
        if only_layers is not None and layer not in only_layers:
            continue
        if name in conversions:
            converted = conversions[name](e, path.vertices).strip()
            if len(converted) > 0:
                # only save if we converted something
                collected.append(converted)
        else:
            log.debug("Entity type %s not exported!", name)

    # join all entities into one string
    entities_str = "\n".join(collected)

    # add in the extents of the document as explicit XYZ lines
    hsub = {f"EXTMIN_{k}": v for k, v in zip("XYZ", np.append(path.bounds[0], 0.0))}
    hsub.update({f"EXTMAX_{k}": v for k, v in zip("XYZ", np.append(path.bounds[1], 0.0))})
    # apply a units flag defaulting to `1`
    hsub["LUNITS"] = _UNITS_TO_DXF.get(path.units, 1)
    # run the format for the header
    sections = [template["header"].format(**hsub).strip()]
    # do the same for entities
    sections.append(template["entities"].format(ENTITIES=entities_str).strip())
    # and the footer
    sections.append(template["footer"].strip())

    # filter out empty sections
    # random whitespace causes AutoCAD to fail to load
    # although Draftsight, LibreCAD, and Inkscape don't care
    # what a giant legacy piece of shit
    # create the joined string blob
    blob = "\n".join(sections).replace(_SAFESPACE, " ")
    # run additional self- checks
    if tol.strict:
        # check that every line pair is (group code, value)
        lines = str.splitlines(str(blob))
        # should be even number of lines
        assert (len(lines) % 2) == 0
        # group codes should all be convertible to int and positive
        assert all(int(i) >= 0 for i in lines[::2])
        # make sure we didn't slip any unicode in there
        blob.encode("ascii")

    return blob


def bulge_to_arcs(lines, bulge, bulge_idx, is_closed=False, metadata=None):
    """
    Polylines can have "vertex bulge" which means the polyline
    has an arc tangent to segments, rather than meeting at a
    vertex.

    From Autodesk reference:
    The bulge is the tangent of one fourth the included
    angle for an arc segment, made negative if the arc
    goes clockwise from the start point to the endpoint.
    A bulge of 0 indicates a straight segment, and a
    bulge of 1 is a semicircle.

    Parameters
    ----------------
    lines : (n, 2) float
      Polyline vertices in order
    bulge : (m,) float
      Vertex bulge value
    bulge_idx : (m,) float
      Which index of lines is bulge associated with
    is_closed : bool
      Is segment closed
    metadata : None, or dict
      Entity metadata to add

    Returns
    ---------------
    vertices : (a, 2) float
      New vertices for poly-arc
    entities : (b,) entities.Entity
      New entities, either line or arc
    """
    # make sure lines are 2D array
    lines = np.asanyarray(lines, dtype=np.float64)

    # make sure inputs are numpy arrays
    bulge = np.asanyarray(bulge, dtype=np.float64)
    bulge_idx = np.asanyarray(bulge_idx, dtype=np.int64)

    # filter out zero- bulged polylines
    ok = np.abs(bulge) > 1e-5
    bulge = bulge[ok]
    bulge_idx = bulge_idx[ok]

    # metadata to apply to new entities
    if metadata is None:
        metadata = {}

    # if there's no bulge, just return the input curve
    if len(bulge) == 0:
        index = np.arange(len(lines))
        # add a single line entity and vertices
        entities = [Line(index, **metadata)]
        return lines, entities

    # use bulge to calculate included angle of the arc
    angle = np.arctan(bulge) * 4.0
    # the indexes making up a bulged segment
    tid = np.column_stack((bulge_idx, bulge_idx - 1))
    # if it's a closed segment modulus to start vertex
    if is_closed:
        tid %= len(lines)

    # the vector connecting the two ends of the arc
    vector = lines[tid[:, 0]] - lines[tid[:, 1]]

    # the length of the connector segment
    length = np.linalg.norm(vector, axis=1)

    # perpendicular vectors by crossing vector with Z
    perp = np.cross(
        np.column_stack((vector, np.zeros(len(vector)))),
        np.ones((len(vector), 3)) * [0, 0, 1],
    )
    # strip the zero Z
    perp = util.unitize(perp[:, :2])

    # midpoint of each line
    midpoint = lines[tid].mean(axis=1)

    # calculate the signed radius of each arc segment
    radius = (length / 2.0) / np.sin(angle / 2.0)

    # offset magnitude to point on arc
    offset = radius - np.cos(angle / 2) * radius

    # convert each arc to three points:
    # start, any point on arc, end
    three = np.column_stack(
        (lines[tid[:, 0]], midpoint + perp * offset.reshape((-1, 1)), lines[tid[:, 1]])
    ).reshape((-1, 3, 2))

    # if we're in strict mode make sure our arcs
    # have the same magnitude as the input data
    if tol.strict:
        from ..arc import arc_center

        check_angle = [arc_center(i).span for i in three]
        assert np.allclose(np.abs(angle), np.abs(check_angle))

        check_radii = [arc_center(i).radius for i in three]
        assert np.allclose(check_radii, np.abs(radius))

    # collect new entities and vertices
    entities, vertices = [], []
    # add the entities for each new arc
    for arc_points in three:
        entities.append(Arc(points=np.arange(3) + len(vertices), **metadata))
        vertices.extend(arc_points)

    # if there are unconsumed line
    # segments add them to drawing
    if (len(lines) - 1) > len(bulge):
        # indexes of line segments
        existing = util.stack_lines(np.arange(len(lines)))
        # remove line segments replaced with arcs
        for line_idx in grouping.boolean_rows(
            existing, np.sort(tid, axis=1), np.setdiff1d
        ):
            # add a single line entity and vertices
            entities.append(Line(points=np.arange(2) + len(vertices), **metadata))
            vertices.extend(lines[line_idx].copy())

    # make sure vertices are clean numpy array
    vertices = np.array(vertices, dtype=np.float64)

    return vertices, entities


def get_key(blob, field, code):
    """
    Given a loaded (n, 2) blob and a field name
    get a value by code.
    """
    try:
        line = blob[np.nonzero(blob[:, 1] == field)[0][0] + 1]
    except IndexError:
        return None
    if line[0] == code:
        try:
            return int(line[1])
        except ValueError:
            return line[1]
    else:
        return None


# store the loaders we have available
_dxf_loaders = {"dxf": load_dxf}



"""
exceptions.py
----------------

Wrap exceptions.
"""


class ExceptionWrapper:
    """
    Create a dummy object which will raise an exception when attributes
    are accessed (i.e. when used as a module) or when called (i.e.
    when used like a function)

    For soft dependencies we want to survive failing to import but
    we would like to raise an appropriate error when the functionality is
    actually requested so the user gets an easily debuggable message.
    """

    def __init__(self, exception: BaseException):
        # store the exception type and the args rather than the whole thing
        # this prevents the locals from the time of the exception
        # from being stored as well.
        self.exception = (type(exception), exception.args)

    def __getattribute__(self, *args, **kwargs):
        # will raise when this object is accessed like an object
        # if it's asking for our class type return None
        # this allows isinstance() checks to not re-raise
        if args[0] == "__class__":
            return None.__class__

        # re-create our original exception from the type and arguments
        exc_type, exc_args = super().__getattribute__("exception")
        raise exc_type(*exc_args)

    def __call__(self, *args, **kwargs):
        # behave the same when this object is called like a function
        # as when someone tries to access an attribute like a module
        self.__getattribute__("exception")
